
#include "ipp.h"
#include "adapt.h"
#include "adj.h"


// The following functions are for action vector managing

BOOLEAN find_act_vect(token name, int level)
{
   action *temp_act;

   for (temp_act=act_vect[level]; temp_act; temp_act=temp_act->next)
   {
      if (strcmp(temp_act->name, name) == SAME)
     	 return (TRUE);
   }

   return (FALSE);
}


// Insert an action, if not present

void insert_act_vect(token name, int level)
{
   action *temp_act ;

   if ( !find_act_vect(name, level) )
   {
      temp_act = ( action* ) calloc( 1, sizeof( action ) );
      temp_act->name = new_token( strlen( name ) + 1 );
      strcpy(temp_act->name, name);
      temp_act->next= act_vect[level];
      act_vect[level]=temp_act;
   }

}


// Free an entire level of action vector

void free_act_vect(int level)
{
   action *temp_act, *temp;

   for(temp_act=act_vect[level]; temp_act; )
   {
      temp = temp_act;
      temp_act = temp->next;
      free(temp);
   }

   act_vect[level]=NULL;
}


// All levels from "level" are moved one step up, when we want to insert a new (longer) plan

void up_act_vect(int level)
{
   int i;

   for (i=adapt.plan_length-1; i>=level; i--)
   {
      act_vect[i+1]=act_vect[i];
      correct_vect[i+1]=correct_vect[i];
   }

   act_vect[level]=NULL;
   adapt.plan_length++;

   if (adapt.goal_agenda)
      cons_update_level(level, 1);
}


// All levels from "level" are moved one step down, when we want to insert a new (shorter) plan

void down_act_vect(int level)
{
   int i;

   if (act_vect[level])
      free_act_vect(level);

   for (i=level; i<adapt.plan_length-1; i++)
   {
      act_vect[i]=act_vect[i+1];
      correct_vect[i]=correct_vect[i+1];
   }

   act_vect[i]=NULL;

   if (adapt.plan_length!=0)
      adapt.plan_length--;

   if (adapt.goal_agenda)
      cons_update_level(level, -1);
}



// Insert new sub-plan

void insert_sub_plan(int init_time, int size)
{
   int i;
   vertex_list op;

   for(i=0; i<size; i++)
   {
      if (act_vect[init_time+i])
         free_act_vect(init_time+i);

      get_next( op_table[i], INIT );/* initialise */
      while ( ( op = get_next( op_table[i], EXEC ) ) != NULL )
	 if(op->is_used && !op->is_noop)
            insert_act_vect(op->name, init_time+i);
   }
}



// Action vector is updated, correction vector is updated too (up or down)

void update_act_vect(int curr_init_time, int curr_max_time, int max_time)
{
   int diff, i;
   int orig_size, new_size;

   orig_size = curr_max_time-curr_init_time;
   new_size = max_time;

   diff = new_size-orig_size;

   if ( diff>0 )
   {
      for (i=0; i<diff; i++)
         up_act_vect(curr_max_time+i);
   }

   else if ( diff<0 )
   {
      diff = -diff;
      for (i=0; i<diff; i++)
	down_act_vect(curr_max_time-i-1);
   }

   insert_sub_plan(curr_init_time, new_size);

}


// Correct vector is used during optimization process

void init_correct_vect(int curr_init_time, int max_time, int try)
{
  int i;

  for(i=0; i<=adapt.plan_length; i++)
     if ( (correct_vect[i]>0) && (correct_vect[i] < try) )
        correct_vect[i]=try;

  for(i=0; i<=max_time; i++)
    correct_vect[curr_init_time+i]=try;
}



void update_correct_vect(int curr_init_time, int curr_max_time, int try)
{
  int i;

  for(i=curr_init_time; i<curr_max_time; i++)
     correct_vect[i]=try;

}



int max_adapt(int max_size, int *max_time, int *result_build)
{

  int result=0;  
     
  for ( ; *max_time <= max_size && !result ; (*max_time)++ ) 
  {   

    if ( display_info > 2 ) 
       printf("\n SEARCH max_time: %d max_size %d \n", *max_time, max_size);  
 
   if ( *result_build && ( result = search_plan( *max_time ) ) != FALSE )   
      break;   
      /* couldn't find plan; check for unsolvability */   
      
    if ( same_as_prev_flag ) 
    {

      /* start after graph has leveled off */   
      if ( previous_Scount == current_Scount ) 
      {   
	/* nothing has changed in last step -> unsolvable */   
	break;   
      }
      else 
      {   
	/* still improvements going on */   
	previous_Scount = current_Scount;   
      }       
    }   
    /* we must continue search with one more layer */          
           
    if( adapt.sub_plan_length <= *max_time )
    {
	build_graph_layer(FALSE);   
	*result_build= are_there_non_exclusive( goal_facts, *max_time);
	if(*result_build==FALSE)
           result= *result_build;
           
	adapt.sub_plan_length++;
    }     
  }

  return (result);  
}  
  


int adapt_build_graph(int *max_time, int max_size)
{   
  struct tms build_s, build_e; 
  int reached_goals, time=0, first=0, minim;        
  same_as_prev_flag=0;
  
  times(&build_s); 

  if (adapt.change==CHANGE_GOAL)   
  {   

      for( minim=1; minim<max_size; minim++ )   
      {   
	  if(minim>=adapt.sub_plan_length)
	  {   
	      build_graph_layer(FALSE);   
	      adapt.sub_plan_length++;
	  }   
	  
	  reached_goals = are_there_non_exclusive( goal_facts, minim );   
	  if ( reached_goals ) break;
	  if ( same_as_prev_flag ) break;   
       }   
       
      *max_time=minim;    
        
  }  
  
  else 
  { 
      adapt.sub_plan_length=0;
      if(fact_table)  
       	free_table(fact_table, MAX_PLAN+1);   
      fact_table=NULL;   
      if(op_table)
        free_table(op_table, MAX_PLAN);   
      op_table=NULL;   

      time=1;
      /* build graph until non-exclusive goals are reached */
      reached_goals = build_graph( &time, FALSE );   
   
      /* plan until goals are reached or graph has leveled off or time>max_size */   
      for( ; time <=max_size ; time++ ) 
      {   
	reached_goals = are_there_non_exclusive( goal_facts, time );
	if ( reached_goals ) 
	{   
           if ( display_info && time > 0 && first )           
              printf("\nADJ: goals first reachable in %d time steps \n", time);
          
  	   break;   
	}   
	if ( same_as_prev_flag ) break;   
	build_graph_layer(FALSE);   
      }
       
      adapt.sub_plan_length=time;
      *max_time = time; 
    } 
 
  times(&build_e); 
 
  build_ad_time += ( float ) ( ( build_e.tms_utime - build_s.tms_utime + 
		        	 build_e.tms_stime - build_s.tms_stime  ) / 100.0 ) ;
 
  return (reached_goals);   
}    



// Detect if initial facts or goals are changed, or both

int detect_plan_changes(int curr_init_time, int curr_max_time, int max_size, int *max_time)
{

   int result_build;
   int goal_time=curr_max_time;


   if (adapt.change==CHANGE_GOAL || adapt.change==TRUE)
   {
      switch (adapt.goal_heuristics)
      {
         case SIMPLE:
         {
            compute_goals_simple(goal_time, adapt.plan_length);
            break;
         }

         case CAUSAL_LINKS:
         {
            compute_CL_goals(goal_time);
            break;
         }
      }

      if (adapt.change==CHANGE_GOAL )
	 result_build=adapt_build_graph(max_time, max_size);			      	
   }   
       
   if(adapt.change==CHANGE_INITIAL || adapt.change==TRUE)    
   { 
      initial_facts=fact_table_vect[curr_init_time];   
      result_build=adapt_build_graph(max_time, max_size);

   }

   if(adapt.change==CHANGE_INITIAL)
     de_alloc(*max_time);

   return (result_build);
}



// Set in the new plan if initial facts or goals are changed, or both

void set_plan_changes(int *curr_init_time, int *curr_max_time)
{

   // we try to expand the planning graph backward, better because no new levels needed  ...

   if ( ((adapt.alterna)%2) <1)
   {
      if (*curr_max_time < adapt.plan_length)
      {
	 (*curr_max_time)++;
	 adapt.change=CHANGE_GOAL;
      }
      else
	 if (*curr_init_time>0)
	 {
            (*curr_init_time)--;   
	    adapt.change=CHANGE_INITIAL;   
	 }
   }

   else
   {
      if (*curr_init_time>0)   
      {  
	 (*curr_init_time)--;   
	 adapt.change=CHANGE_INITIAL;   
      }
      else   
	 if (*curr_max_time < adapt.plan_length)   
	 { 
	    (*curr_max_time)++;     
	    adapt.change=CHANGE_GOAL;   
	 }
   }

   adapt.alterna++;
}


void set_initial_times(int *curr_init_time, int *curr_max_time, int *initial_max_size)
{
  
   if (*curr_init_time<adapt.plan_length)
   {
      (*curr_max_time)=(*curr_init_time)+1;

      // empty levels ...
      while (act_vect[*curr_max_time]==NULL && *curr_max_time<adapt.plan_length)
      {
         (*curr_max_time)++;
         (*initial_max_size)++;
      }	                 
      
   }      

   else
   {
      *curr_max_time=*curr_init_time;
      (*curr_init_time)--;
      
      // empty levels ...
      while (act_vect[*curr_init_time]==NULL && *curr_init_time>0)
      {
         (*curr_init_time)--;
         (*initial_max_size)++;
      }
					          
   }
}


// Set current initial time and current max time
// Used during plan optimization process

void set_current_times(int *curr_init_time, int *curr_max_time, int try)
{
   int i;

   for(i=0; i<adapt.plan_length && correct_vect[i]==0; i++);
   (*curr_init_time)=i;

   for(; i<adapt.plan_length && correct_vect[i]==try; i++);
   (*curr_max_time)=i;

   if ( (*curr_max_time)==adapt.plan_length && (*curr_init_time)!=0 )
   {
      (*curr_init_time)--;
   }
   else
      if( (*curr_init_time)==0 && (*curr_max_time)!=adapt.plan_length )
      {
         (*curr_max_time)++;
      }
}



/* the setting up function  called by main
 * simply sets up search and does the initial call
 * as well as memoizing the result if search was a failure
 */

BOOLEAN adapt_plan( int *curr_init_time_pointer )
{

  int max_size=MAX_SIZE;
  int result=FALSE, result_build;
  int max_time=1, curr_max_time, i;
  int curr_init_time = *curr_init_time_pointer;

  adapt.task_type = SEARCH;
  adapt.step = STEP;

  expand(adapt.plan_length+20);    /* AAAAAAAAAAAAAAAAA */

  if (adapt.goal_agenda)
  {
     i = first_reordered_goal();

     if ( i != INVALID_VALUE && i < curr_init_time && i < adapt.plan_length-1 )
     {
        curr_init_time = i;
	adapt.num_inc = 1; // goal reordering
     }
  }

  max_size = MAX_SIZE + adapt.num_inc -1;

  set_initial_times( &curr_init_time, &curr_max_time, &max_size);

  same_as_prev_flag=0;
  adapt.change=TRUE;

  // we search for a plan of "max_size" maximum length starting from level "curr_init_time"

  while( (result==FALSE || result==PARTIAL_TIMEOUT) && adapt.sub_plan_length < MAX_PLAN )
  {

     adapt.try++;
     result_build=detect_plan_changes(curr_init_time, curr_max_time, max_size, &max_time);

     printf("\n Replanning Window <%d, %d>", curr_init_time, curr_max_time);
     printf("  (iteration: %d  try: %d  max_size: %d)\n", adapt.iteraz, adapt.try, max_size);

     adapt.change=FALSE;

     result=max_adapt(max_size, &max_time, &result_build);

     if(result==TRUE)
     {
        adapt.step=STEP;

        if (adapt.goal_agenda)
           cons_reset_goal_flag(curr_init_time+max_time);

        printf("\n\nPLAN WINDOW TO ADAPT\n");
        print_facts_and_goals(curr_init_time, curr_max_time);

        update_act_vect(curr_init_time, curr_max_time, max_time);
        init_correct_vect(curr_init_time, max_time, adapt.try);

        printf("\nPARTIAL SOLUTION (time %d)\n\n",curr_init_time);
        print_plan(curr_init_time, curr_init_time+max_time);

        de_alloc(max_time);
     }

     else if (result == TIMEOUT)
        break;

     else
     {
        set_plan_changes(&curr_init_time, &curr_max_time);
        max_size+=adapt.step;
	adapt.step++;
     }
  }

  (*curr_init_time_pointer) = curr_init_time;
  return (result);
}



// Compute plan optimization (more compact solutions)

int resize_plan(char* fact_file_name)
{

  int max_size, result, result_build;
  int max_time=1,curr_init_time, curr_max_time;

  int max_size_count=0;
  BOOLEAN stop=FALSE;

  float plan_time;
  struct tms end_time;

  adapt.best_plan_length = adapt.plan_length;
  adapt.best_num_actions = adapt.num_actions;
  adapt.task_type = OPTIMIZE;

  expand(adapt.plan_length+20); /* AAAAAAAAAAAAAAAAA */

  printf("\nCOMPUTING ADDITIONAL MORE COMPACT SOLUTIONS\n");


  while ( !stop )
  {

      adapt.iteraz++;
      set_current_times(&curr_init_time, &curr_max_time, adapt.try);

      same_as_prev_flag=0;
      adapt.sub_plan_length=0;
      adapt.change=TRUE;
      result=FALSE;

      // we search for a (better) plan of "max_size" maximum length starting from level "curr_init_time"

      while ( (result==FALSE || result==PARTIAL_TIMEOUT) && adapt.sub_plan_length < MAX_PLAN && !stop )
      {

        adapt.try++;

	max_size= curr_max_time-curr_init_time-1;
        result_build=detect_plan_changes(curr_init_time, curr_max_time, max_size, &max_time);

	adapt.change=FALSE;
        printf("\n Replanning Window <%d, %d> ",curr_init_time,curr_max_time);
        printf("  (iteration: %d  try: %d  max_size: %d)\n", adapt.iteraz, adapt.try, max_size);

        result=max_adapt(max_size, &max_time, &result_build);

        times(&end_time);

	plan_time = ( float ) ( ( end_time.tms_utime - start_time.tms_utime +
 		                  end_time.tms_stime - start_time.tms_stime ) / 100.0 ) ;

	if ( result==TRUE )
	{

	   printf("\n\nPLAN WINDOW TO OPTIMIZE (%d)\n", adapt.iteraz);
           print_facts_and_goals(curr_init_time, curr_max_time);

  	   update_act_vect(curr_init_time, curr_max_time, max_time);
           update_correct_vect(curr_init_time, curr_max_time, adapt.try);

	   if ( display_info > 1 )
	   {
    	      printf("\nPARTIAL SOLUTION\n\n");
              print_plan(curr_init_time, curr_init_time+max_time);

	   }

  	   de_alloc(max_time);

           plan_check();

 	   if ( adapt.plan_length < adapt.best_plan_length ||
              ( adapt.plan_length==adapt.best_plan_length && adapt.num_actions < adapt.best_num_actions ))
           {

              printf("\nOPTIMIZED PLAN (%d)\n\n", adapt.iteraz);
              print_plan(0, adapt.plan_length);
              print_adj_report(plan_time);
              store_adapted_plan( adapt.plan_length, fact_file_name, plan_time );
              fflush(stdout);

              // this is only for testing
              #ifdef TEST_ADJ
                 printf("\nAABBCC%d::%.2f::%d::%d\n", adapt.iteraz, plan_time, adapt.plan_length, adapt.num_actions );
//                 exit(0);
              #endif /* TEST_ADJ */

 	      adapt.best_plan_length = adapt.plan_length;
              adapt.best_num_actions = adapt.num_actions;
 	   }

  	   else
	   {
	      printf("\nADJ: the previous plan is already optimized ! \n\n");
              stop=TRUE;
	   }

	}

	else if ( result==TIMEOUT )
        {
           printf("\nADJ: timeout reached, please increase time limit! \n");
	   return(result);
	}	
	
	else
	   set_plan_changes(&curr_init_time, &curr_max_time);
		
	
        if ((curr_init_time<=0) && (curr_max_time>=adapt.plan_length))
        {
	   max_size_count++;	// iteration on the last replanning window
	
  	   if ( max_size_count>=2 )
 	      stop=TRUE;
        }
 	
 	if (stop)
 	{
           printf("\n\nBEST PLAN\n\n"); 	
           print_plan(0, adapt.plan_length);
           print_adj_report(plan_time);

           // this is only for testing
           #ifdef TEST_ADJ
              printf("\nAABBCCO::%.2f::%d::%d\n", plan_time, adapt.plan_length, adapt.num_actions );
           #endif /* TEST_ADJ */
        }
			
      }

  }

  return (result);
}


void de_alloc(int max_time)
{

  int i;

  choose_critics(max_time, CLEAR);

  for ( i=0; i < max_time; i++ )
  {
     num_goals_at[i]=0;
     num_ops_at[i]=0;
     num_SS_at[i]=0;
     num_critics_at[i]=0;
  }

//  choose_critics(max_time, CLEAR);
}


void free_table(hashtable *t, int max_size)

{
  int i, j;
  vertex_list v, v1;

  for(i=0;i<max_size;i++)
    if(t[i]!=NULL)
      for(j=0; j<HSIZE; j++)
      {
	for( v = t[i][j]; v!=NULL; )
	{
	  v1=v;
	  v = v->next;
	  free_vertex(v1);
	}
	t[i][j]=NULL;
      }
  free(t);
}


void free_edge_list(edge_list e)
{
  edge_list e_l;
  for(; e ;)
  {
     e_l=e;
     e=e->next;
     free(e_l);
  }
}


void free_cond_edge_list(cond_edge_list ce)
{
  cond_edge_list ce_l;
  for(; ce ;)
  {
     ce_l=ce;
     ce=ce->next;
     if(ce_l->conditions)
        free_edge_list(ce_l->conditions);

     free(ce_l);
  }
}


void free_delete_list(delete_list de)
{
  delete_list de_l;
  for(; de ;)
  {
     de_l=de;
     de=de->next;
     if(de_l->conditions)
	free_edge_list(de_l->conditions);

     free(de_l);
  }
}


void free_vertex(vertex_list v)
{

  if(v->name)
    free(v->name);
  /*    else return; */
  v->name=NULL;
  if( v->precond_edges )
    free_edge_list( v->precond_edges );
  v->precond_edges=NULL;
  if( v->add_edges)
    free_cond_edge_list( v->add_edges);
  v->add_edges=NULL;
  if( v->del_list)
    free_delete_list( v->del_list);
  v->del_list=NULL;
  if(v->memo_start)
    free_sub_tree(v->memo_start);
  v->memo_start=NULL;

  free(v);

}


// these are for reading/writing plan files

FILE* open_plan_file(char* file_name)
{
  FILE* fp;

  fp = fopen(file_name,"r");

  if(fp==NULL)
  {
    printf("\nADJ: can't find plan file %s ! \n\n", file_name);
    exit (2);
  }

  return (fp);
}


int test_level(char* cStr)
{
  int i;

  if(cStr[strlen(cStr)-1] == ':')
     i = 1;
  else
     i = 0;

  return(i);
}


int estract_level(char* cStr)
{
  int n,m;

  n = strlen(cStr) - 1;
  cStr[n]='\n';
  m=atoi(cStr);

  return(m);
}



// Search for comment in the input file : they starts whith '#'

void skip_comment(FILE* fp )
{
  char charact;

  charact = fgetc(fp);

  if (charact == '#')
  {
     while ( charact != '\n' )
        charact = fgetc(fp);
  }

}


int init_act_vect(FILE* fp)
{
  int level;
  char cStr[256];

  while (!feof(fp))
  {
     skip_comment(fp);
     fscanf(fp,"%s",cStr);

     if (test_level(cStr))
     {
       level = estract_level(cStr);
       fscanf(fp,"%s",cStr);
     }

     insert_act_vect(cStr, level);
  }

  return(level+1);
}


// Load input plan file

int load_plan(char* plan_file)
{
  int plan_length;
  FILE* pFile;

  pFile = open_plan_file(plan_file);
  plan_length=init_act_vect(pFile);
  fclose(pFile);

  return(plan_length);
}


// Store adapted plan on file

void store_adapted_plan(int levels, char* fct_file_name, double time)
{

  int i;
  char cNameFile[256];
  static int num_plan=0;
  FILE* fp;

  action *temp_act;

  num_plan++;

  sprintf(cNameFile,"plan_%s_%d.SOL\0", fct_file_name, num_plan);
  if((fp = fopen( cNameFile, "w")) == NULL)
  {
      printf("\nERROR: cannot open file ! \n\n");
      exit(1);
  }
  
  fprintf(fp,"#Problem:\t%s\t time:\t%f\t levels:\t%d\n", fct_file_name, time, levels);

  /* for all non empty operator levels */
  for ( i=0; i<levels; i++ )
  {
     fprintf(fp,"%3d: ", i);

     /* for all operators in plan */
     for(temp_act = act_vect[i]; temp_act; temp_act = temp_act->next)
     {
        fprintf(fp,"\t%s\n", temp_act->name );
     }

  }

  fclose(fp);
}


