
#include "ipp.h" 
#include "adapt.h"
#include "adj.h"


// Initialization for adj and fixpoint

void initialize_adj_fixpoint()
{
  int i;
 
  for ( i = 0; i<G_HSIZE; i++ ) 
  {
     g_table[i] = NULL;
     o_table[i] = NULL;
  }

  incr_goal_facts = NULL;    

  gam_total_num_ops = 0;
  gam_total_num_goals = 0;

  gam_fixpoint_reached = FALSE;
  gam_fixpoint_level = 0;

  bytes_o = 0;
  bytes_f = 0;
  bytes_e = 0;
  bytes_k = 0;
  bytes_excl = 0;
  numbereff = 0;
  numberkan = 0;
  numberexclkan = 0;

  bytes_graph = 0;
  bytes_memo = 0;

}


// Clean g_table variables used in forward propagation

static void init_g_table()
{
    register int i;
    register g_node_list go;
    
    for ( i = 0; i < G_HSIZE; i++ ) 
      for ( go = g_table[i]; go; go = go->next ) 
      {
      
	  go->adj.is_true = FALSE;
 	  go->adj.is_precond = FALSE;
	  go->adj.is_added = FALSE;
	  go->adj.is_deleted = FALSE;
          go->adj.is_goal = FALSE;

      }
	    
}


// Clean o_table variables used in ineffective actions remotion

static void init_o_table()
{
    register int i;
    register o_node_list op;
    
    for ( i = 0; i < G_HSIZE; i++ ) 
      for ( op = o_table[i]; op; op = op->next ) 
      {
	  op->is_used = FALSE;
      }

}


// Print critic informations about a goal node

void print_goal(g_node_list go, BOOLEAN details)
{
   printf("\n - %s ",go->name);

   if (details)
   {
      printf(" (");  
      printf(" is_true: %d ", go->adj.is_true);
      printf(" is_precond: %d ", go->adj.is_precond);
      printf(" is_added: %d ", go->adj.is_added);
      printf(" is_deleted: %d ", go->adj.is_deleted);
      printf(" is_goal: %d ", go->adj.is_goal);
      printf(")\n");
   }
}



// Print critic informations about an operator node

void print_operator(o_node_list op, BOOLEAN details)
{
 
   printf("\n - %s ",op->name);

   if (details)
   {
      printf(" (");  
      printf(" is_used: %d ", op->is_used);
      printf(")");   
   }

} 


void print_g_table()
{

   int i;
   g_node_list go;
   
   BOOLEAN goal_details = TRUE;
   
   printf("\n Contenuto della g_table :\n");

   for ( i = 0; i < G_HSIZE; i++ )
     for ( go = g_table[i]; go; go = go->next ) 
         print_goal(go,goal_details);

   printf("\n");
       
}


void print_o_table()
{

   int i;
   o_node_list op;
   
   BOOLEAN operators_details = TRUE;
   
   printf("\n Contenuto della o_table :\n");
     
   for ( i = 0; i < G_HSIZE; i++ ) 
     for ( op = o_table[i]; op; op = op->next ) 
        print_operator(op,operators_details);
   
   printf("\n");
       
}



// Set initial true facts at "time_init"

void set_initial_true_facts(int time_init)
{

  int i;
  token_list temp_ft, initial;
  g_node_list go;
  
  BOOLEAN goal_details = FALSE;
    
  initial = fact_table_vect[time_init];
  /* initial facts if time_init = 0 */
      
  for( temp_ft = initial; temp_ft; temp_ft = temp_ft->next )   
  {  
     
     go = lookup_go ( temp_ft->item );
     
     if ( go==NULL )
     {
        printf("\n ADJ: Initial fact '%s' not found! Debug me.\n", temp_ft->item);  
        continue;
     }
     else
        go->adj.is_true = TRUE;
  }

  if ( debug == 3 )
  {
     printf("\n Initial facts at level %d : \n", time_init);    
     
     for ( i = 0; i < G_HSIZE; i++ ) 
        for ( go = g_table[i]; go; go = go->next )     
        {
           if ( go->adj.is_true )
               print_goal(go,goal_details);
        }
        
     printf("\n");
  }

}



// Search and count the inconsistences in g_table

int incons_check(int level)
{
 
  int i;
  int num_incons= NO_INCONS;
  
  BOOLEAN goal_details = TRUE;
  
  g_node_list go; 

  for ( i = 0; i < G_HSIZE; i++ ) 
     for ( go = g_table[i]; go; go = go->next ) 
     {

        if ((go->adj.is_precond && !go->adj.is_true) ||
            (go->adj.is_added && go->adj.is_deleted))
        {
	      
	   if ( num_incons == NO_INCONS )  /* first inconsistence found */
              num_incons = 1;
	   else 
	      num_incons++;   
	      
	   printf("\n ADJ: Inconsistent fact '%s' at level %d ", go->name, level);
	
	   if ( debug == 3 )
	   {
 	      print_goal(go,goal_details);
              printf("\n");
           }	
        }
     }

  if ( debug == 3 ) printf("\n");
  return (num_incons);
}



// Count not satisfied goals at level "time"

int goal_check(int time)
{

  int not_satisfied = 0;

  token_list temp_go;
  g_node_list go;

  for( temp_go = store_goal_facts; temp_go; temp_go = temp_go->next )
  {
    go = lookup_go ( temp_go->item );
    if ( go->adj.is_true == FALSE )
    {
       printf("\n ADJ: Goal '%s' not satisfied at level %d ", go->name, time);
       not_satisfied++;
    }
  }

  return(not_satisfied);
}



// Set preconditions at level "time"

void set_preconditions(int time)
{
  
  action *temp_act;    
            
  o_node_list an_action;
  g_edge_list prec;
    
  for(temp_act = act_vect[time]; temp_act; temp_act = temp_act->next)
  {
     an_action = lookup_op(temp_act->name);
  
     if ( an_action==NULL )
     {
        if ( debug == 3 ) 
           printf("\n ADJ: Action '%s' not found ! Time %d \n", temp_act->name, time);
        continue;
     }

     for (prec = an_action->preconds; prec; prec = prec->next)  
     {                
        prec->endpt->adj.is_precond = TRUE;
     }
     
  }
  
}
     


// Apply actions of plan at level "time" : set action effects

int apply_actions(int time)
{

  int num_actions=0;

  action *temp_act;    

  o_node_list an_action;
  e_node_list an_effect;
  g_edge_list add_effect, del_effect;

  for(temp_act = act_vect[time]; temp_act; temp_act = temp_act->next)
  {
     an_action = lookup_op(temp_act->name);
  
     if ( an_action==NULL )
     {
        if ( debug == 3 )
           printf("\n ADJ: Action '%s' not found ! \n", temp_act->name);

        continue;	
     }
     else
        num_actions++;

     for (an_effect = an_action->effects; an_effect ; an_effect = an_effect->next)
     {
  
     /* conditional effects */
        	  
        if (an_effect->conditions != NULL)
        {
           printf("\n ADJ: Error, unable to use conditional effects !\n");
           exit(1);
        }
        
                   
     /* positive effects */
     	
	for (add_effect = an_effect->add_edges; add_effect ; add_effect = add_effect->next)
	{
           add_effect->endpt->adj.is_added = TRUE;
        } 
	
     /* negative effects */
     	
        for (del_effect = an_effect->del_edges; del_effect ; del_effect = del_effect->next)
        { 	   
 	   del_effect->endpt->adj.is_deleted = TRUE; 	    	
        }

     }

  }

  return (num_actions);
  
}



// Update facts F(i)

void update_facts(int time)
{

  int i;
  token_list temp;
  g_node_list go;

  /* fact_table_vect[time] is freed and re-writed whith new facts */
  free_token_list(fact_table_vect[time]);
  fact_table_vect[time] = NULL;
  
  /* we identificate new true facts */
  /* positive and negative effects are inserted, other g_table variables are cleaned */

  for ( i = 0; i < G_HSIZE; i++ ) 
    for ( go = g_table[i]; go; go = go->next )
    {  
    
       if (go->adj.is_added)
       {
           go->adj.is_true = TRUE;
           go->adj.is_added = FALSE;
           
       }
       if (go->adj.is_deleted)
       {
           go->adj.is_true = FALSE;
           go->adj.is_deleted = FALSE;
          
       }
       if (go->adj.is_true)
       {
   	   temp =  ( token_list ) calloc( 1, sizeof( token_list_elt ) );   
           temp->item = new_token( strlen( go->name ) + 1 );
       	   strcpy(temp->item, go->name);   
           temp->next = fact_table_vect[time];
           fact_table_vect[time]=temp;         
       }
   
       go->adj.is_precond = FALSE;
                
    }

  if ( debug == 3 )
  {
     printf("\n True facts at level %d : \n", time); 
     print_list(fact_table_vect[time]);     
  }
     
}



// Forward propagation of facts : it founds inconsistences and counts the actions applyed

int forward_propagation(int time_init, int time_end)
{

  int incons_level = NO_INCONS;
  int num_incons = NO_INCONS;

  int i, not_satisfied_goals;

  init_g_table();

  printf("\n Starting forward propagation from level %d to %d ... \n", time_init, time_end);

  set_initial_true_facts(time_init);


  for ( i = time_init; i < time_end ; i++)
  {

     set_preconditions(i);
     adapt.num_actions += apply_actions(i);
     num_incons = incons_check(i);

     if ( num_incons > 0 )
     {
        incons_level = i;
        break;
     }

     else
        update_facts(i+1);

  }

  /* goals check for last level */

  if ( i==adapt.plan_length )
  {

     not_satisfied_goals = goal_check(i);

     if ( not_satisfied_goals > 0 )
     {
        num_incons = not_satisfied_goals;
        incons_level = time_end;
     }

  }

  if (num_incons > 0)
     printf("\n\n ADJ: Found %d inconsistence/s at time %d \n\n", num_incons, incons_level);     
      	    	
  adapt.num_inc = num_incons;
  return(incons_level);

}



// Causal links
// Find causal links that matches input mask

causal_link_list find_causal_link(token fact_name, token name_from, int level_from,
                                  token name_to, int level_to, BOOLEAN initialize)
{

   static causal_link_list initial_pointer;

   causal_link_list temp_link;

   if ( initialize )
   {
      initial_pointer = causal_links;
      return (NULL);
   }


   for(temp_link=initial_pointer; temp_link; temp_link=temp_link->next)
   {

      if ( ( ( strcmp(temp_link->from.action, name_from) == SAME ) ||
             ( strcmp(name_from, ALL_NAMES) == SAME ) ) &&

           ( ( temp_link->from.time == level_from ) ||
             ( level_from == ALL_LEVELS ) ) &&

           ( ( strcmp(temp_link->to.action, name_to) == SAME ) ||
             ( strcmp(name_to, ALL_NAMES) == SAME ) ) &&

           ( ( temp_link->to.time == level_to ) ||
             ( level_to == ALL_LEVELS ) ) &&

           ( ( strcmp(temp_link->fact, fact_name) == SAME ) ||
             ( strcmp(fact_name, ALL_NAMES) == SAME ) ) )
      {
	 initial_pointer = temp_link->next;
         return (temp_link);
      }

   }

   return (NULL);
}


// Insert a new causal link in list

void insert_causal_link(token fact_name, token name_from, int level_from, token name_to, int level_to)
{

   causal_link_list temp_link;

   temp_link=find_causal_link(fact_name, name_from, level_from, name_to, level_to, INIT);

   if ( (temp_link=find_causal_link(fact_name, name_from, level_from, name_to, level_to, EXEC)) == NULL )
   {

      temp_link = ( causal_link_list ) calloc( 1, sizeof( causal_link_list_elt ) );

      temp_link->from.action = new_token( strlen( name_from ) + 1 );
      strcpy(temp_link->from.action, name_from);
      temp_link->from.time = level_from;
      temp_link->to.action = new_token( strlen( name_to ) + 1 );
      strcpy(temp_link->to.action, name_to);
      temp_link->to.time = level_to;
      temp_link->fact = new_token( strlen( fact_name ) + 1 );
      strcpy(temp_link->fact, fact_name);

      // first element inserted
      if (causal_links == NULL)
      {
         temp_link->previous = NULL;
         temp_link->next = NULL;
         causal_links = temp_link;
      }

      else
      {
         temp_link->previous = NULL;
         causal_links->previous = temp_link; 	
         temp_link->next = causal_links;
         causal_links = temp_link;
      }

   }

}


// Remove a causal link

void remove_causal_link(token fact_name, token name_from, int level_from, token name_to, int level_to)
{

   causal_link_list temp_link;

   temp_link=find_causal_link(fact_name, name_from, level_from, name_to, level_to, INIT);

   while ( (temp_link=find_causal_link(fact_name, name_from, level_from, name_to, level_to, EXEC)) != NULL )
   {
      temp_link->previous->next = temp_link->next;
      temp_link->next->previous = temp_link->previous;
      free(temp_link);
   }

}


// Clean causal links list : links are re-calculated every time

void free_causal_links()
{

   causal_link_list temp_link, temp;

   for(temp_link=causal_links; temp_link; )
   {
      temp = temp_link;
      temp_link = temp->next;
      free(temp);
   }

   causal_links = NULL;
}


void print_link(causal_link_list temp_link)
{
   printf("\n action1 : %s",temp_link->from.action);
   printf("\n time1   : %d",temp_link->from.time);
   printf("\n fact    : %s",temp_link->fact);
   printf("\n action2 : %s",temp_link->to.action);
   printf("\n time2   : %d\n",temp_link->to.time);
}


void print_causal_links()
{
   int i=0;
   causal_link_list temp_link;

   printf("\n CAUSAL LINKS\n");

   for(temp_link=causal_links; temp_link; temp_link=temp_link->next)
   {
      i++;
      printf("\n %d)",i);
      print_link(temp_link);
   }

}


// Find a couple action & time in a CL list

action_at_time_list find_CL(action_at_time_list initial_pointer, token act_name, int level)
{

   action_at_time_list temp_CL;

   for (temp_CL=initial_pointer; temp_CL; temp_CL=temp_CL->next)
   {
      if ( ( ( strcmp(temp_CL->of.action, act_name) == SAME ) ||
             ( strcmp(temp_CL->of.action, ALL_NAMES) == SAME ) ) &&

           ( ( temp_CL->of.time == level ) ||
             ( temp_CL->of.time == ALL_LEVELS ) ) )

      return (temp_CL);
   }

   return (NULL);

}


// Insert a couple action & time in a CL list

action_at_time_list insert_CL(action_at_time_list initial_pointer, token act_name, int level)
{

   action_at_time_list temp_CL;

   if ( !find_CL(initial_pointer, act_name, level) )
   {
      temp_CL = ( action_at_time_list ) calloc( 1, sizeof( action_at_time_list_elt ) );

      temp_CL->of.action = new_token( strlen( act_name ) + 1 );
      strcpy(temp_CL->of.action, act_name);
      temp_CL->of.time = level;

      temp_CL->next = initial_pointer;
      initial_pointer = temp_CL;
   }

   return (initial_pointer);
}


// Clean a CL list

void free_CL(action_at_time_list initial_pointer)
{

   action_at_time_list temp_CL, temp;

   for (temp_CL=initial_pointer; temp_CL; )
   {
      temp = temp_CL;
      temp_CL = temp->next;
      free (temp);
   }
}


// Clean all CL lists of a fact

void free_CL_lists(g_node_list fact)
{
   free_CL(fact->adj.precond);
   fact->adj.precond = NULL;
   fact->adj.is_precond = FALSE;
   free_CL(fact->adj.add_effect);
   fact->adj.add_effect = NULL;
   fact->adj.is_added = FALSE;
   free_CL(fact->adj.del_effect);
   fact->adj.del_effect = NULL;
   fact->adj.is_deleted = FALSE;
}


// Clean all CL lists for all facts in g_table

void free_all_CL_lists()
{
   int i;
   g_node_list fact;

   for ( i = 0; i < G_HSIZE; i++ )
      for ( fact = g_table[i]; fact; fact = fact->next )
         free_CL_lists(fact);
}


void print_CL(action_at_time_list initial_pointer)
{
   int i=0;
   action_at_time_list temp_CL;

   for(temp_CL=initial_pointer; temp_CL; temp_CL=temp_CL->next)
   {
      i++;
      printf("\n   %d) %s (time %d) \n", i, temp_CL->of.action, temp_CL->of.time);
   }
}


void print_goal_CL(g_node_list go)
{
   printf("\n\n   DEST. ACTIONS (preconditions list)\n");
   print_CL(go->adj.precond);
   printf("\n   SOURCE ACTIONS (add effects list)\n");
   print_CL(go->adj.add_effect);
}


// Remove not_applicable  actions from the input plan

void remove_not_applicable_actions(int max_time)
{

   action *curr_act, *prec_act=NULL;
   o_node_list an_action;
   int time;

   printf("\n Not Applicable Actions: \n");

   for (time=0; time<max_time; time++)
   {
      for (curr_act=act_vect[time]; curr_act; )
      {

	 an_action = lookup_op(curr_act->name);
	
	 if ( an_action == NULL )
	 {
	     // action is removed if not used
	     printf("\n - Action '%s' removed from plan at level %d ", curr_act->name , time);

	     // first element
	     if ( curr_act == act_vect[time] )
	     {
		 act_vect[time] = curr_act->next;
		 free(curr_act);
		 prec_act=NULL;
		 curr_act=act_vect[time];
	     }
	
	     // element in a generic position
	     else
	     {
		 prec_act->next = curr_act->next;
		 free(curr_act);
                 curr_act = prec_act->next;
	     }
	 }
	 else
	 {
  	    prec_act=curr_act;
	    curr_act=curr_act->next;
	}
     }
     
  }
}



// Remove an action from level "time" of action vector, if not used

BOOLEAN remove_ineffective_actions(int time)
{

   BOOLEAN action_removed=FALSE;
   action *curr_act, *prec_act=NULL;
   o_node_list an_action;

   for (curr_act=act_vect[time]; curr_act; )
   {

      an_action = lookup_op(curr_act->name);

      if ( an_action != NULL )
      {
         if ( !an_action->is_used )
         {

	    action_removed=TRUE;
            // action is removed if not used
            printf("\n - Action '%s' removed from plan at level %d ", an_action->name, time);

            // first element
            if ( curr_act == act_vect[time] )
            {
               act_vect[time] = curr_act->next;
               free(curr_act);
               prec_act=NULL;
               curr_act=act_vect[time];
            }

            // element in a generic position
            else
            {
               prec_act->next = curr_act->next;
               free(curr_act);
               curr_act = prec_act->next;
            }
         }

         // action is used
         else
         {
            prec_act=curr_act;
            curr_act=curr_act->next;
            an_action->is_used = FALSE; // reset is_used
         }
      }
      else
      {
         prec_act=curr_act;
         curr_act=curr_act->next;
      }
   }

   return (action_removed);
}


// Build causal links for level "time"

void check_for_links_at_level(time)
{

   int i;
   g_node_list fact;
   o_node_list an_action;

   action_at_time_list temp_precond, temp_add_effect;

   for ( i = 0; i < G_HSIZE; i++ )
      for ( fact = g_table[i]; fact; fact = fact->next )
      {

         if (!fact->adj.is_deleted)
         {

            if (fact->adj.is_precond && fact->adj.is_added)
            {
               for (temp_precond=fact->adj.precond; temp_precond; temp_precond=temp_precond->next)
                  for (temp_add_effect=fact->adj.add_effect; temp_add_effect; temp_add_effect=temp_add_effect->next)
                  {
                     insert_causal_link(fact->name, temp_add_effect->of.action, temp_add_effect->of.time,
                                        temp_precond->of.action, temp_precond->of.time);

                     an_action=lookup_op(temp_add_effect->of.action);

                     if ( an_action!=NULL )
                        an_action->is_used = TRUE;
                        // if an action has no effect in  the plan is marked as not used

                  }

               free_CL(fact->adj.add_effect);
               fact->adj.add_effect = NULL;
               fact->adj.is_added = FALSE;
            }
         }

         else // if (fact->adj.is_deleted)
            free_CL_lists(fact);

      }

}


// Compute all causal links examining the actual plan starting from goals level

void compute_causal_links(token_list actual_goal_facts)
{

  int time;
  int init_act_time= -1;

  action *temp_act;
  token_list temp_fact;

  o_node_list an_action;
  e_node_list an_effect;
  g_node_list a_fact;
  g_edge_list prec, add_effect, del_effect;

  BOOLEAN temp_action_removed, action_removed=FALSE;

  free_causal_links();
  free_all_CL_lists();
  init_o_table();

  printf("\n Computing causal links from level 0 to %d ... \n", adapt.plan_length);

  // setting goals as preconditions of GOAL_ACT
  for (temp_fact=actual_goal_facts; temp_fact; temp_fact=temp_fact->next)
  {
     a_fact = lookup_go (temp_fact->item);

     if ( a_fact == NULL )
     {
        if ( debug == 3 ) 
           printf("\n ADJ: Goal fact '%s' not found ! \n", temp_fact->item);
        continue;
     }

     a_fact->adj.precond=insert_CL(a_fact->adj.precond, GOAL_ACT, adapt.plan_length);
     a_fact->adj.is_precond = TRUE;
  }


  for (time=adapt.plan_length-1; time>=0; time--)
  {

     for(temp_act = act_vect[time]; temp_act; temp_act = temp_act->next)
     {
        an_action = lookup_op(temp_act->name);

        if ( an_action==NULL )
        {
           if ( debug == 3 )
              printf("\n ADJ: Action '%s' not found ! \n", temp_act->name);

           continue;
        }

        for (an_effect = an_action->effects; an_effect ; an_effect = an_effect->next)
        {

        /* conditional effects */

           if (an_effect->conditions != NULL)
           {
              printf("\n ADJ: Error, unable to use conditional effects !\n");
              exit(1);
           }


        /* positive effects */

	   for (add_effect = an_effect->add_edges; add_effect ; add_effect = add_effect->next)
    	   {
              a_fact = add_effect->endpt;
              a_fact->adj.add_effect=insert_CL(a_fact->adj.add_effect, an_action->name, time);
	      a_fact->adj.is_added = TRUE;
           }

        /* negative effects */

           for (del_effect = an_effect->del_edges; del_effect ; del_effect = del_effect->next)
           {
 	      a_fact = del_effect->endpt;
 	      a_fact->adj.del_effect=insert_CL(a_fact->adj.del_effect, an_action->name, time);
    	      a_fact->adj.is_deleted = TRUE;
           }
        }
     }

     check_for_links_at_level(time);

     if (adapt.action_remotion)
     {
        temp_action_removed=remove_ineffective_actions(time);
	if (action_removed==FALSE)
	   action_removed=temp_action_removed;
     }

     for (temp_act=act_vect[time]; temp_act; temp_act=temp_act->next)
     {
        an_action = lookup_op(temp_act->name);

        if ( an_action!=NULL )
	{
           for (prec = an_action->preconds; prec; prec = prec->next)
           {
              a_fact = prec->endpt;
              a_fact->adj.precond=insert_CL(a_fact->adj.precond, an_action->name, time);
              a_fact->adj.is_precond = TRUE;
           }
        }
     }
  }

  // setting initial facts as (positive) effects of INIT_ACT
  for (temp_fact=store_initial_facts; temp_fact; temp_fact=temp_fact->next)
  {
     a_fact = lookup_go (temp_fact->item);
     a_fact->adj.add_effect=insert_CL(a_fact->adj.add_effect, INIT_ACT, init_act_time);
     a_fact->adj.is_added = TRUE;
  }

  check_for_links_at_level(time);

  if ( debug > 4 )
     print_causal_links();

  if ( action_removed ) printf("\n");
}


// Search all causal links involved in goals calculus : at level "goal_time"

void extract_actual_causal_links(int goal_time)
{

   causal_link_list temp_link;
   g_node_list a_fact;

   free_all_CL_lists();

   for(temp_link=causal_links; temp_link; temp_link=temp_link->next)
   {

      if ( ( temp_link->from.time < goal_time ) && ( temp_link->to.time >= goal_time ) )
      {
         if (debug == 4)
            print_link(temp_link);

         a_fact = lookup_go (temp_link->fact);
         a_fact->adj.precond=insert_CL(a_fact->adj.precond, temp_link->to.action, temp_link->to.time);
         a_fact->adj.is_precond = TRUE;
         a_fact->adj.add_effect=insert_CL(a_fact->adj.add_effect, temp_link->from.action, temp_link->from.time);
         a_fact->adj.is_added = TRUE;
      }
   }

}


// Analize a causal link : count causal links, find nearest and last link (in time)

void analize_causal_link(action_at_time_list initial_pointer, int *links_count,
                         int *prox_link_level, int *last_link_level)
{

   action_at_time_list temp_CL;

   *links_count=0;
   *prox_link_level=adapt.plan_length;
   *last_link_level=0;

   for (temp_CL=initial_pointer; temp_CL; temp_CL=temp_CL->next)
   {
      (*links_count)++;
      if (temp_CL->of.time < *prox_link_level)
         *prox_link_level = temp_CL->of.time;
      if (temp_CL->of.time > *last_link_level)
         *last_link_level = temp_CL->of.time;
   }

}


// Set values "new_value" for temporary deleted goals

void set_temp_deleted_goals_as(BOOLEAN new_value)
{
  int j;
  g_node_list go;

  for ( j = 0; j < G_HSIZE; j++ )
    for ( go = g_table[j]; go; go = go->next )
    {
       if ( go->adj.is_goal == TEMP_DELETED )
          go->adj.is_goal = new_value;
    }

}



// Find couples of mutex goals

g_node_list find_mutex_goals(g_node_list go, BOOLEAN initialize)
{

  int j;
  static int initial_index;
  static g_node_list initial_pointer;

  g_node_list temp_go;

  BOOLEAN goal_details = FALSE;


  if ( initialize )
  {
     initial_index = 0;
     initial_pointer = g_table[initial_index];
     return (NULL);
  }

  for ( j = initial_index; j < G_HSIZE; )
  {
     for ( temp_go = initial_pointer; temp_go; temp_go = temp_go->next )
     {
        if (temp_go->adj.is_goal)
        {
           if ( gam_go_are_mutex(go,temp_go,gam_fixpoint_level) )
           {

              if ( debug == 4 )
              {
                 printf("\n\n Goals are mutex :\n");
                 print_goal(go,goal_details);
                 print_goal_CL(go);
                 print_goal(temp_go,goal_details);
                 print_goal_CL(temp_go);
                 printf("\n");
              }

              initial_index = j;
              initial_pointer = temp_go->next;

              return(temp_go);
           }
        }
     }

     j++;
     initial_pointer = g_table[j];

  }

  return(NULL);

}


// Set as initial sub-goals the preconditions of actions at the next level

void set_initial_goals_simple(void)
{

  int j;
  g_node_list go;

  for ( j = 0; j < G_HSIZE; j++ )
    for ( go = g_table[j]; go; go = go->next )
    {

       if (go->adj.is_precond)
       {
          go->adj.is_goal = TRUE;
          go->adj.is_precond = FALSE;
       }
    }
}


/* setta is_goal e resetta is_precond per tutti i goal il cui val
ore   nella cons_list e' minore o uguale al goal_time passato */

void set_initial_goals(int goal_time)
{

  goal_set_list st;
  goal_list gl;
  g_node_list go;

  for ( st=cons_start->right; st!=cons_finish && st->level < goal_time; st=st->right )
    for ( gl = st->set; gl!=NULL; gl = gl->next )
    {

       go=lookup_go(gl->name);
       go->adj.is_goal = G_ORD_GOAL;
       go->adj.is_precond = FALSE;

       if ( debug == 6 )
	  printf("\n G_ORD : %s  %d ",go->name,go->adj.is_goal);
    }
}


// Apply goals as actions at last level

void apply_goals()
{

  token_list temp_ft;
  g_node_list go;


  for( temp_ft = store_goal_facts; temp_ft; temp_ft = temp_ft->next )
  {

     go = lookup_go ( temp_ft->item );

     if ( go==NULL )
     {
        printf("\n ADJ: Goal fact '%s' not found! Debug me.\n", temp_ft->item);
        continue;
     }
     else
        go->adj.is_precond = TRUE;
  }
}



// Verify if a goal is mutex whith other goals

BOOLEAN goal_is_mutex(g_node_list go)
{

  int j;
  g_node_list temp_go;

  BOOLEAN goal_details = FALSE;

  for ( j = 0; j < G_HSIZE; j++ )
     for ( temp_go = g_table[j]; temp_go; temp_go = temp_go->next )
     {
        if (temp_go->adj.is_goal)
        {
           if ( gam_go_are_mutex(go,temp_go,gam_fixpoint_level) )
           {

              if ( debug == 4 )
              {
                 printf("\n\n Goals are mutex :\n");
                 print_goal(go,goal_details);
                 print_goal(temp_go,goal_details);
                 printf("\n");
              }

              return(TRUE);
           }
        }
     }

  return(FALSE);

}


// Update goals G(j)

void update_goals_simple()
{

  int j;
  g_node_list go;

  /* we identificate new goals */

  for ( j = 0; j < G_HSIZE; j++ )
    for ( go = g_table[j]; go; go = go->next )
    {

       if ( go->adj.is_precond && !go->adj.is_goal )
       {
          if ( !( go->adj.is_added || go->adj.is_deleted || goal_is_mutex(go) ) )
          {
             go->adj.is_goal = TRUE;
             go->adj.is_precond = FALSE;

          }

       }

    }
}



// Compute new sub-goals applying actions or goals, if last level reached

void compute_goals_simple(int goal_time, int time_end)
{

   int j;
   token_list temp;
   g_node_list go;

   init_g_table();

   if ( goal_time != adapt.plan_length )
   {

      printf("\n Searching for goals at level %d ... \n", goal_time);

      set_preconditions(goal_time);
      set_initial_goals_simple();
      apply_actions(goal_time);

      for ( j = goal_time+1; j <=time_end; j++ )
      {

         if ( j != adapt.plan_length )
         {
            set_preconditions(j);
            update_goals_simple();
            apply_actions(j);
         }

         else
         {
            apply_goals();
            update_goals_simple();
         }

      }


      /* goal_facts is freed and re-writed whith new goals */

      free_token_list(goal_facts);
      goal_facts = NULL;

      for ( j = 0; j < G_HSIZE; j++ )
        for ( go = g_table[j]; go; go = go->next )
        {

           if (go->adj.is_goal)
           {
 	      temp =  ( token_list ) calloc( 1, sizeof( token_list_elt ) );
              temp->item = new_token( strlen( go->name ) + 1 );
       	      strcpy(temp->item, go->name);
              temp->next = goal_facts;
              goal_facts = temp;
           }

        }

   }

   else
   {
      goal_facts = copy_token_list( store_goal_facts );

   }

   if (debug == 3)
   {
      printf("\n\n Goals at level %d : \n", goal_time);
      print_list(goal_facts);
   }

}



/* Update goals G(j) using causal links heuristics */

void update_CL_goals(int goal_time)
{

  int j;
  int links_count, prox_link_level, last_link_level;
  int m_links_count, m_prox_link_level, m_last_link_level;

  g_node_list go, mutex_go;

  BOOLEAN new_goal_is_best;
  BOOLEAN add_condition = TRUE;  // if goal agenda is not used

  /* we identificate new goals */

  for ( j = 0; j < G_HSIZE; j++ )
    for ( go = g_table[j]; go; go = go->next )
    {

       new_goal_is_best = TRUE;

       if (adapt.goal_agenda)
          add_condition = (adapt.task_type==SEARCH) || !is_already_goal_at_time(go->name, goal_time);

       if ( go->adj.is_precond && go->adj.is_added && add_condition )
       {

          mutex_go=find_mutex_goals(go, INIT);

          while ( (mutex_go=find_mutex_goals(go, EXEC)) != NULL )
          {

             if (adapt.goal_agenda)
             {
          	if ( (adapt.task_type==SEARCH) && mutex_go->adj.is_goal == G_ORD_GOAL )
          	 {
          	  new_goal_is_best = FALSE;
          	  break;
          	 }
             }

             analize_causal_link(go->adj.precond, &links_count, &prox_link_level, &last_link_level);
             analize_causal_link(mutex_go->adj.precond, &m_links_count, &m_prox_link_level, &m_last_link_level);

             // switch causal links heuristics ....

             if ( ( links_count < m_links_count ) || ( m_links_count == 0 ) ||
                  ( (links_count == m_links_count) && (prox_link_level < m_prox_link_level) ) )
             {
                new_goal_is_best = TRUE;
                mutex_go->adj.is_goal = TEMP_DELETED;
                // if new goal is better than others, the others are temporary deleted
             }

             else
             {
                new_goal_is_best = FALSE;
                break;
             }
          }

          if ( new_goal_is_best )
          {
             go->adj.is_goal = TRUE;
             go->adj.is_precond = FALSE;
             go->adj.is_added = FALSE;

             set_temp_deleted_goals_as(FALSE);
             // worst goals are deleted from goals list
          }

          else
             set_temp_deleted_goals_as(TRUE);
             // temporary deleted goals are re-writed in goals list

       }

    }

}


// Compute new sub-goals applying causal links heuristics

void compute_CL_goals(int goal_time)
{

   int j;
   token_list temp;
   g_node_list go;

   init_g_table();

   if ( goal_time != adapt.plan_length )
   {

      printf("\n Searching for goals at level %d ... \n", goal_time);

      extract_actual_causal_links(goal_time);

      if (adapt.goal_agenda)
      {
         if (adapt.task_type==SEARCH)
            set_initial_goals(goal_time);
      }

      update_CL_goals(goal_time);

      /* goal_facts is freed and re-writed whith new goals */

      free_token_list(goal_facts);
      goal_facts = NULL;

      for ( j = 0; j < G_HSIZE; j++ )
        for ( go = g_table[j]; go; go = go->next )
        {

           if (go->adj.is_goal)
           {
	      temp =  ( token_list ) calloc( 1, sizeof( token_list_elt ) );
              temp->item = new_token( strlen( go->name ) + 1 );
       	      strcpy(temp->item, go->name);
              temp->next = goal_facts;
              goal_facts = temp;
           }

        }

   }

   else
   {
      goal_facts = copy_token_list( store_goal_facts );
   }

   if (debug == 3)
   {
      printf("\n\n Goals at level %d : \n", goal_time);
      print_list(goal_facts);
   }

}


// removes empty steps in the plan

void remove_empty_steps(int max_time)
{
  int time;

  for(time=0; time<max_time; )
  {
    if ( act_vect[time]==NULL )
    {
       down_act_vect(time);
       max_time--;
    }
    else
       time++;
  }
}


// Final check on founded plan, it uses forward propagation

void plan_check()
{

  int incons_level;

  printf("\n Checking the plan ... \n");

  printf("\n Removing empty steps ... \n");
  remove_empty_steps(adapt.plan_length);

  if ( adapt.goal_heuristics == CAUSAL_LINKS )
     compute_causal_links(store_goal_facts);

  adapt.num_actions=0;
  incons_level = forward_propagation(0, adapt.plan_length);

  if (incons_level != NO_INCONS)
  {
    printf("\n ERROR : Found plan is NOT VALID !!!\n\n");
    exit(1);
  }
  else
    printf("\n Ok \n\n");

}


void print_plan(int time_init, int time_end)
{

  int i;          /* index for time */
  BOOLEAN first;  /* for better printing: is it the first one this level? */

  action *temp_act;

  /* for all non empty operator levels */
  for ( i=time_init; i<time_end; )
  {

     printf("time step %3d: ", i);
     first = TRUE;

     if (act_vect[i]==NULL)
        printf("\t%s\n", "(no actions)");

     /* for all operators in plan */
     for(temp_act = act_vect[i]; temp_act; temp_act = temp_act->next)
     {
        if ( first )
        {
           printf("\t%s\n", temp_act->name );
           first = FALSE;
        }
        else
	   printf("\t\t%s\n", temp_act->name );
     }

     i++;
  }

  printf( "\n" );

}


void print_facts_and_goals(int curr_init_time, int curr_max_time)
{

   printf("\nINITIAL FACTS   (time %d) \n", curr_init_time);
   print_list(initial_facts);

   printf("\nGOALS           (time %d) \n", curr_max_time);
   print_list(goal_facts);
   printf("\n");
}


// Printing of a token list : for video output

void print_list(token_list list)
{

  token_list i_token;

  for ( i_token = list; i_token; i_token = i_token->next )
    printf( "\n - %s ", i_token->item );

  printf("\n");

}
