/*
 * this one builds the fix point level of the planning graph
 * into a slightly different ( i.e. for gam better ) data structure,
 * in order to achieve exclusivity information.
 */


#include"ipp.h"
#include"adj.h"


int build_fixpoint( void )

{

  token_list t;
  int time = 0;
  int gam_old_excl_sum, gam_excl_sum = 0;
  operator_list o;
  g_node_list go;

  for ( t = initial_facts; t; t = t->next ) 
  {
    go = insert_go( t->item );
    go->is_initial = TRUE;
  }


  while ( !gam_fixpoint_reached )
  {

    if ( time == MAX_PLAN ) 
    {
      printf("\ngam: didn't have time to find fixpoint.increase MAX_PLAN\n\n");
      exit( 1 );
    }

    gam_fixpoint_reached = TRUE;

    insert_noops();
  
    for ( o = operators; o; o = o->next ) 
    {
      gam_apply_operator( o, time );
    }

    gam_find_mutex_ops_and_insert_deletes( time );

    gam_old_excl_sum = gam_excl_sum;
    gam_excl_sum = gam_find_mutex_facts( time + 1 );
  
    /* print out info, if required */
    if ( display_info ) 
    {
      printf( "gam time: %3d, %3d facts, %3d ops and %5d exclusive pairs\n",
	      time+1, gam_total_num_goals, gam_total_num_ops, gam_excl_sum );
      if ( display_info > 1 )
	printf( "gam memory: %.1f Kbytes graph, %.1f Kbytes exclusives\n",
		(float) ( (bytes_o + bytes_f + bytes_e + bytes_k) / 1024),
		(float) (bytes_excl / 1024) );
    }
    
    if ( gam_old_excl_sum == gam_excl_sum  && gam_fixpoint_reached ) 
    {
      gam_fixpoint_level = time;
      
      if ( display_info ) 
        printf( "\ngam: graph fix point reached at time step %d\n", time+1 );

    } 
    else 
    {
      gam_fixpoint_reached = FALSE;
    }
  
    /* increment time */
    time++;
    
  }

  if ( display_info > 1 ) 
  {
    printf("a total number of %d facts, %d ops with %ld effects connected by %ld edges\n", 
            gam_total_num_goals, gam_total_num_ops, numbereff, numberkan - numberexclkan);
    
    printf("Kbytes: %.1f on ops, %.1f facts, %.1f effects, %.1f edges\n",
	   (float) bytes_o/1024 ,
	   (float) bytes_f/1024 ,
	   (float) bytes_e/1024 ,
	   (float) bytes_k/1024);
  }
  
  return (gam_fixpoint_level);
}


void gam_apply_operator( operator_list o, int time )

{ 

  static g_array_type g_pre, g_con;/* array for precond facts */

  int num_pre, num_con, i, j;

  o_node_list o_node;

  inst_effect_list e;/* index */
  g_node_list go;
  e_node_list t_e;

  token_list t;

  /* check if the preconds are there
   * and load them into ft array
   */
  if ( !gam_get_them_non_exclusive( o->preconditions, time, &g_pre, &num_pre ) ) 
     return;

  if ( ( o_node = lookup_op( o->name ) ) != NULL ) 
  {
    for ( e = o->effects; e; e = e->next ) 
    {
      if ( ( t_e = effect_is_already_there( o_node, e, time ) ) != NULL ) 
      {
	t_e->is_new = FALSE;
	continue;
      }
      if ( gam_get_them_non_exclusive( e->conditions, time, &g_con, &num_con ) ) 
      {
	for ( i=0; i<num_pre; i++ ) 
	{
	  for ( j=0; j<num_con; j++ ) 
	  {
	    if ( gam_go_are_mutex( g_pre[i], g_con[j], time ) )
	      break;
	  }
	  if ( j<num_con ) break;
	}
	if ( i<num_pre ) continue;

	t_e = ( e_node_list ) calloc( 1, sizeof( e_node_list_elt ) );
	CHECK_MEMORY( t_e ); GRAPHE( e_node_list_elt );
	t_e->on = TRUE;
	t_e->is_new = TRUE;
	t_e->del_names = e->del_effects; 
	t_e->op = o_node;
	t_e->conditions = NULL;
	for ( i=0; i<num_con; i++ ) 
	{
	  t_e->conditions = insert_g_edge( g_con[i], t_e->conditions );
	  g_con[i]->effconds = insert_e_edge( t_e, g_con[i]->effconds );
	}
	t_e->add_edges = NULL;
	for ( t = e->add_effects; t; t = t->next ) 
	{
	  if ( ( go = lookup_go( t->item ) ) == NULL ) 
	  {
	    gam_fixpoint_reached = FALSE;
	    go = insert_go( t->item );
	  }
	  t_e->add_edges = insert_g_edge( go, t_e->add_edges );
	  go->add_edges = insert_e_edge( t_e, go->add_edges );
	}
	t_e->next = o_node->effects;
	o_node->effects = t_e;
      }
    }

    return;
  }

  gam_fixpoint_reached = FALSE;

  /* apply operator,
   * first get new vertex 
   */
  o_node = insert_op( o->name );

  /* connect the preconds */
  for ( i=0; i<num_pre; i++ ) 
  {
    o_node->preconds = insert_g_edge( g_pre[i], o_node->preconds );
    g_pre[i]->preconds = insert_o_edge( o_node, g_pre[i]->preconds );
  }

  /* insert add edges */
  o_node->effects = NULL;
  for ( e = o->effects; e; e = e->next ) 
  {
    if ( gam_get_them_non_exclusive( e->conditions, time, &g_con, &num_con ) ) 
    {
      for ( i=0; i<num_pre; i++ ) 
      {
	for ( j=0; j<num_con; j++ ) 
	{
	  if ( gam_go_are_mutex( g_pre[i], g_con[j], time ) )
	    break;
	}
	if ( j<num_con ) break;
      }
      if ( i<num_pre ) continue;

      t_e = ( e_node_list ) calloc( 1, sizeof( e_node_list_elt ) );
      CHECK_MEMORY( t_e ); GRAPHE(e_node_list_elt);
      t_e->on = TRUE;
      t_e->is_new = TRUE;
      t_e->del_names = e->del_effects; 
      t_e->op = o_node;
      t_e->conditions = NULL;
      for ( i=0; i<num_con; i++ ) 
      {
	t_e->conditions = insert_g_edge( g_con[i], t_e->conditions );
	g_con[i]->effconds = insert_e_edge( t_e, g_con[i]->effconds );
      }
      t_e->add_edges = NULL;
      for ( t = e->add_effects; t; t = t->next ) 
      {
	if ( ( go = lookup_go( t->item ) ) == NULL ) 
	{
	  go = insert_go( t->item );
	}
	t_e->add_edges = insert_g_edge( go, t_e->add_edges );
	go->add_edges = insert_e_edge( t_e, go->add_edges );
      }
      t_e->next = o_node->effects;
      o_node->effects = t_e;
    }
  }

}


void gam_find_mutex_ops_and_insert_deletes( int time )

{

  o_node_list o, o_;
  g_node_list g;
  e_node_list e;
  g_edge_list i_ge, j_ge;
  o_edge_list i_oe;
  e_edge_list i_ee;
  token_list t;
  o_excl_list temp;
  int i, j;

  for ( i=0; i<G_HSIZE; i++ ) 
  {
    for ( o = o_table[i]; o; o = o->next ) 
    { 
      temp = ( o_excl_list ) calloc ( 1, sizeof( o_excl_list_elt ) );
      CHECK_MEMORY(temp); EXCL(o_excl_list_elt); 
      o->excl_at[time] = temp;
    }
  }

  for ( i=0; i<G_HSIZE; i++ ) 
  {
    for ( o = o_table[i]; o; o = o->next ) 
    { 

      for ( j = 0; j < G_HSIZE; j++ ) 
      {
	for ( o_ = o_table[j]; o_; o_ = o_->next ) 
	{
	  if ( ( int ) o < ( int ) o_ ) continue;
	  for ( i_ge = o->preconds; i_ge; i_ge = i_ge->next ) 
	  {
	    for ( j_ge = o_->preconds; j_ge; j_ge = j_ge->next ) 
	      if ( gam_go_are_mutex( i_ge->endpt, j_ge->endpt, time ) ) break;
	    if ( j_ge ) break;
	  }
	  if ( i_ge )
	    gam_op_make_exclusive( o, o_, time );
	}
      }
      

      for ( e = o->effects; e; e = e->next ) 
      {

	if ( e->is_new ) 
	{
	  for ( t = e->del_names; t; t = t->next ) 
	  {
	    g = lookup_go( t->item );
	    if ( !g ) continue;

	    e->del_edges = insert_g_edge( g, e->del_edges );
	    g->del_edges = insert_e_edge( e, g->del_edges );
	  }
	}

	if ( e->conditions ) continue;
	
	for ( i_ge = e->del_edges; i_ge; i_ge = i_ge->next ) 
	{
	  g = i_ge->endpt;
 
	  for ( i_oe = g->preconds; i_oe; i_oe = i_oe->next ) 
	  { 
	    if ( o != i_oe->endpt )
	      gam_op_make_exclusive( o, i_oe->endpt, time );
	  }

	  for ( i_ee = g->add_edges; i_ee; i_ee = i_ee->next ) 
	  {
	    if ( ( !(i_ee->endpt->conditions) ) && ( i_ee->endpt->op != o ) )
	      gam_op_make_exclusive( o, i_ee->endpt->op, time );
	  }
	}
      }

    }
  }

}


int gam_find_mutex_facts( int time )

{

  static BOOLEAN initialise = TRUE;
  static g_node_list *g;

  int i, j, num = 0, count = 0;
  g_node_list g_;
  g_excl_list temp;

  if ( initialise ) 
  {
    g = ( g_node_list * ) calloc( max_nodes, sizeof( g_node_list ) );
    CHECK_MEMORY( g ); 
    bytes_excl += max_nodes * sizeof( g_node_list );
    initialise = FALSE;
  }

  for ( i = 0; i<G_HSIZE; i++ ) 
  {
    for ( g_ = g_table[i]; g_; g_ = g_->next ) 
    {
       g[num++] = g_;

       temp = ( g_excl_list ) calloc ( 1, sizeof( g_excl_list_elt ) );
       CHECK_MEMORY(temp);
       EXCL(g_excl_list_elt);
       g_->excl_at[time] = temp;
    }
  }

  for ( i=0; i < num; i++ ) 
  {
    for ( j=i+1; j<num; j++ ) 
    {
      if ( !gam_go_are_mutex( g[i], g[j], time ) ) 
      {
	if ( gam_facts_are_exclusive( g[i], g[j], time ) ) 
	{
	  count++;
	  gam_go_make_exclusive( g[i], g[j], time );
	}
      }
    }
  }

  return count;

}


BOOLEAN gam_facts_are_exclusive( g_node_list g1, g_node_list g2, int time )

{

  e_edge_list i_ee, j_ee;
  g_edge_list i_ge, j_ge;
  o_node_list o1, o2;
  

  for ( i_ee = g1->add_edges; i_ee; i_ee = i_ee->next ) 
  {
    for ( j_ee = g2->add_edges; j_ee; j_ee = j_ee->next ) 
    {
      o1 = i_ee->endpt->op;
      o2 = j_ee->endpt->op;
      if ( gam_op_are_mutex( o1, o2, time - 1 ) ) continue;
      
      for ( i_ge = i_ee->endpt->conditions; i_ge; i_ge = i_ge->next ) 
      {
	for ( j_ge = j_ee->endpt->conditions; j_ge; j_ge = j_ge->next ) 
	{
	  if ( gam_go_are_mutex( i_ge->endpt, j_ge->endpt, time-1 ) ) break;
	}
	if ( j_ge ) break;
	for ( j_ge = o2->preconds; j_ge; j_ge = j_ge->next ) 
	{
	  if ( gam_go_are_mutex( i_ge->endpt, j_ge->endpt, time-1 ) ) break;
	}
	if ( j_ge ) break;
      }
      if ( i_ge ) continue;
      
      for ( i_ge = o1->preconds; i_ge; i_ge = i_ge->next ) 
      {
	for ( j_ge = j_ee->endpt->conditions; j_ge; j_ge = j_ge->next ) 
	{
	  if ( gam_go_are_mutex( i_ge->endpt, j_ge->endpt, time-1 ) ) break;
	}
	if ( j_ge ) break;
      }
      if ( i_ge ) continue;

      if ( debug == 7 ) 
      {
	printf("\nfound good pair %s and %s for %s and %s at %d\n",
	       o1->name, o2->name, g1->name, g2->name, time);
      }

      return FALSE;

    }
  }

  return TRUE;

}


void insert_noops( void )

{

  int i;
  g_node_list g;
  o_node_list o;
  e_node_list temp;

  for ( i=0; i<G_HSIZE; i++ ) 
  {
    for ( g = g_table[i]; g; g = g->next ) 
    {
      g->is_new = FALSE;
      if ( !g->has_noop ) 
      {
	o = insert_op( make_noop_string( g->name ) );

	temp = ( e_node_list ) calloc( 1, sizeof( e_node_list_elt ) );
	CHECK_MEMORY( temp ); 
	GRAPHO(e_node_list_elt);
	temp->conditions = NULL;
	temp->del_edges = NULL;
	temp->add_edges = NULL;
	temp->add_edges = insert_g_edge( g, temp->add_edges );
	temp->is_new = FALSE;
	temp->del_names = NULL;
	temp->op = o;
	temp->on = FALSE;

	g->add_edges = insert_e_edge( temp, g->add_edges );

	temp->next = o->effects;
	o->effects = temp;

	o->preconds = NULL;
	o->preconds = insert_g_edge( g, o->preconds );
	g->preconds = insert_o_edge( o, g->preconds );

	o->is_noop = TRUE;
	g->has_noop = TRUE;
      }
    }
  }

}

