
#include "ipp.h"
#include "adj.h"


void make_gam_goal_facts( goal_set_list list, BOOLEAN first_try )
{

  token_list temp1, temp2;
  goal_list g;

  goal_facts = incr_goal_facts;

  if( first_try ) 
  {
    for ( g = list->set; g; g = g->next ) 
    {
      temp1 = ( token_list ) calloc( 1, sizeof( token_list_elt ) );
      temp1->item = g->name;
      temp1->next = goal_facts;
      goal_facts = temp1;
      if ( !list->intermediates ) 
      {
	temp2 = ( token_list ) calloc( 1, sizeof( token_list_elt ) );
	temp2->item = g->name;
	temp2->next = incr_goal_facts;
	incr_goal_facts = temp2;
      }
    }
  }
  
  if ( list->right->intermediates && !linearize ) 
  {
    if ( !first_try ) 
    {
      printf("\ngam: second trying goal set with intermediates right of it! debug me.\n");
      exit( 1 );
    }
      
    for ( g = list->right->set; g; g = g->next ) 
    {
      temp1 = ( token_list ) calloc( 1, sizeof( token_list_elt ) );
      temp1->item = g->name;
      temp1->next = goal_facts;
      goal_facts = temp1;
    }
  }
  
}


void setup_gam_initial_facts( int time )
{

  int i;
  g_node_list g;
  token_list temp;

  initial_facts = NULL;

  for ( i = 0; i < time; i++ )
    evaluate( i );
    
  for ( i = 0; i < G_HSIZE; i++ )
    for ( g = g_table[i]; g; g = g->next ) 
    {
      if ( !g->is_initial ) continue;
      temp = ( token_list ) calloc( 1, sizeof( token_list_elt ) );
      temp->item = g->name;
      temp->next = initial_facts;
      initial_facts = temp;
    }

}


void evaluate( int time )
{

  int i, j, N, count = 0;
  vertex_list op;
  cond_edge_list c_e, c_f;
  edge_list e;
  g_node_list g;

  for( i = 0; i < num_ops_at[time]; i++ )
    if ( !ops_at[time][i]->is_noop ) count++; 
  if ( count <= 1 )
  {
    for ( j = 0; j < G_HSIZE; j++ )
      for ( g = g_table[j]; g; g = g->next )
	g->curr_true = g->is_initial;
    simple_evaluate( time );
    for ( i = 0; i < G_HSIZE; i++ )
      for ( g = g_table[i]; g; g = g->next ) 
      {
	g->is_initial = ( g->initial_count == 1 );
	g->initial_count = 0;
      }
    return;
  }

  if ( 0 ) 
  {
    printf("num ops at time %d is: %d\n", time, num_ops_at[time]);
    for( i = 0; i < num_ops_at[time]; i++ ) printf( ops_at[time][i]->name );
  }

  for ( i = 0; i < num_ops_at[time]; i++ ) 
  {
    op = ops_at[time][i];
    for ( c_e = op->del_edges; c_e; c_e = c_e->next )
    {
      for ( e = c_e->conditions; e; e = e->next )
      {
	for ( c_f = e->endpt->next_time->add_edges; c_f; c_f = c_f->next )
	  if ( c_f->endpt->is_used && c_f->endpt != op && !c_f->endpt->is_noop ) break;
	if ( c_f ) break;
      }
      if ( e ) break;
    }
    if ( c_e ) break;
  }

  for ( j = 0; j < G_HSIZE; j++ )
    for ( g = g_table[j]; g; g = g->next )
      g->curr_true = g->is_initial;

  if ( i < num_ops_at[time] ) 
  {
    N = fac( num_ops_at[time] );
    if ( debug )
      printf("\ngam: using all permutations (%d possibilities ) of parallel set now\n", N);
    permut_evaluate( time );
  } 
  else 
  {
    simple_evaluate( time );
    N = 1;
  }

  for ( i = 0; i < G_HSIZE; i++ )
    for ( g = g_table[i]; g; g = g->next ) 
    {
      g->is_initial = ( g->initial_count == N );
      g->initial_count = 0;
    }

}
    

void permut_evaluate( int time )
{
  simple_evaluate( time );
  permutate( time, 1 );
}


void permutate( int time, int n )
{
 
  int i, j, k, m = num_ops_at[time];
  vertex_list help;

  if ( n == ( m - 1 ) ) 
  {
    help=ops_at[time][m-1];
    ops_at[time][m-1]=ops_at[time][m-2];
    ops_at[time][m-2]=help;
    simple_evaluate( time );
  } 
  else 
  {
    for(i=n+1; i<=m; i++) 
    {
      permutate( time, n + 1 );
      help=ops_at[time][n-1];
      ops_at[time][n-1]=ops_at[time][m+n-i];
      ops_at[time][m+n-i]=help;
      k=m;
      j=n+1;
      while (k>j) 
      {
        help=ops_at[time][k-1];
	ops_at[time][k-1]=ops_at[time][j-1];
	ops_at[time][j-1]=help;
	j++;
	k--;
      }
      simple_evaluate( time );
    }
    permutate( time, n + 1 );
  }
  
}


void simple_evaluate( int time )
{

  int i;
  g_node_list g;
      
  for ( i = 0; i < num_ops_at[time]; i++ )
     op_evaluate( ops_at[time][i] );

  for ( i = 0; i < G_HSIZE; i++ )
    for ( g = g_table[i]; g; g = g->next )
    {
      if ( g->curr_true ) g->initial_count++;
      g->curr_true = FALSE;
    }

}


void op_evaluate( vertex_list op )
{

  o_node_list o;
  e_node_list e;
  g_edge_list g_e;
  g_node_list g;
  int i;
      
  if ( op->is_noop ) return;

  o = lookup_op( op->name );
  if ( !o ) 
  {
    printf("\ndidn't find op where it should be. debug me, please\n\n");
    exit( 1 );
  }
  for ( g_e = o->preconds; g_e; g_e = g_e->next )
    if ( !g_e->endpt->curr_true ) break;
  if ( g_e ) return;

  for ( e = o->effects; e; e = e->next ) 
  {
    for ( g_e = e->conditions; g_e; g_e = g_e->next )
      if ( !g_e->endpt->curr_true ) break;
    if ( g_e ) continue;
    for ( g_e = e->add_edges; g_e; g_e = g_e->next )
      g_e->endpt->made_true = TRUE;
    for ( g_e = e->del_edges; g_e; g_e = g_e->next )
      g_e->endpt->made_false = TRUE;
  }

  for ( i = 0; i < G_HSIZE; i++ )
    for ( g = g_table[i]; g; g = g->next ) 
    {
      if ( g->made_true && g->made_false ) 
      {
	printf("\ngam: op %s has contradictory effects on %s", o->name, g->name);
	exit( 1 );
      }
      if ( g->made_true ) g->curr_true = TRUE;
      if ( g->made_false ) g->curr_true = FALSE;
      g->made_true = FALSE;
      g->made_false = FALSE;
    }
}


void linearize_subplan( goal_set_list old )
{

  goal_set_list temp, left = old->left;
  goal_list g = old->set, help;

  while( g ) 
  {
    temp = ( goal_set_list ) calloc( 1, sizeof( goal_set_list_elt ) );
    temp->left = left;
    left->right = temp;
    left = temp;
    temp->intermediates = old->intermediates;
    help = g->next;
    g->next = NULL;
    temp->set = g;
    g = help;
  }

  temp->right = old->right;
  old->right->left = temp;

}    
 

int fac( int N )
{

  int i, res;

  res = 1;
  for ( i = 2; i <= N; i++ )
     res *= i;

  return (res);

}



token_list gam_token_list_from_token( token str )
{

  char *p = str, *q, temp[MAX_LENGTH];
  token_list_elt dummy, *current;
  current = &dummy;

  while ( *p ) 
  {
    current->next = ( token_list ) calloc( 1, sizeof( token_list_elt ));
    current = current->next;
    for( q = temp; (*p != '\0') && (*p != CONNECTORCHAR); *q++ = *p++);
    if ( *p == CONNECTORCHAR ) ++p;
    *q = '\0';
    current->item = ( char * ) calloc( 1 + strlen( temp ), sizeof( char ));
    CHECK_MEMORY( current->item );
    strcpy(current->item, temp);
  }
  current->next = NULL;
  return ( dummy.next );

}


fact_list fact_list_from_token_list( token_list t )
{

  fact_list result;

  if ( !t ) return NULL;

  result = ( fact_list ) calloc( 1, sizeof( fact_list_elt ) );
  CHECK_MEMORY( result );
  result->item = gam_token_list_from_token( t->item );
  result->next = fact_list_from_token_list( t->next );

  return result;

}


void initialise_set_list( token_list goal_facts )
{

  goal_set_list g;
  fact_list goals_as_facts, i_f;
  token_list i_t;
  goal_list temp;

  goals_as_facts = fact_list_from_token_list( goal_facts );

  g = ( goal_set_list ) calloc( 1, sizeof( goal_set_list_elt ) );
  CHECK_MEMORY( g );
  start = ( goal_set_list ) calloc( 1, sizeof( goal_set_list_elt ) );
  CHECK_MEMORY( start );
  end = ( goal_set_list ) calloc( 1, sizeof( goal_set_list_elt ) );
  CHECK_MEMORY( end );

  start->intermediates = FALSE;
  end->intermediates = FALSE;

  start->right = g;
  g->right = end;
  end->right = NULL;
  start->left = NULL;
  g->left = start;
  end->left = g;

  start->set = NULL;
  end->set = NULL;
  g->set = NULL;

  i_f = goals_as_facts;
  for ( i_t = goal_facts; i_t; i_t = i_t->next ) 
  {
    temp = ( goal_list ) calloc( 1, sizeof( goal_list_elt ) );
    CHECK_MEMORY( temp );
    temp->name = i_t->item;
    temp->objects = i_f->item->next;
    temp->next = g->set;
    g->set = temp;
    i_f = i_f->next;
  }
}


void print_gam_set_list( void )
{

  int t = 1;
  goal_set_list i;
  goal_list j;
  BOOLEAN first;

  for ( i = start->right; i != end; i = i->right ) 
  {
    printf( "step %2d: ", t );
    first = TRUE;
    for ( j = i->set; j; j = j->next ) 
    {
      if ( first ) 
      {
	printf( "%s\n", j->name ); 
	first = FALSE;
	
      } 
      else 
	printf( "\t\t\t%s\n", j->name );
    }
    t++;
  }
}


BOOLEAN oversize( goal_list goals )
{

  token_list so_far = NULL, temp, i, o;
  int n = 0;
  goal_list g;

  for ( g = goals; g; g = g->next )
  {
    for ( o = g->objects ; o; o = o->next )
    {
      for ( i = so_far; i; i = i->next )
	if ( strcmp( i->item, o->item ) == SAME ) break;
      if ( i ) continue;
      temp = ( token_list ) calloc( 1, sizeof( token_list_elt ) );
      CHECK_MEMORY( temp );
      temp->item = o->item;
      temp->next = so_far;
      so_far = temp;
      n++;
      if ( n > max_small_set ) break;
    }
  }

  if ( debug == 2 && n > max_small_set ) 
  {
    printf( "found oversized goal set:\n" );
    for ( g = goals; g; g = g->next ) printf( "%s\n", g->name );
  }

  return ( n > max_small_set );

}


void insert( goal_set_list lp, goal_set_list rp, goal_list goals )
{

  goal_set_list temp;

  temp = ( goal_set_list ) calloc( 1, sizeof( goal_set_list_elt ) );
  CHECK_MEMORY( temp );
  temp->intermediates = TRUE;
  temp->set = goals;
  temp->left = lp;
  temp->right = rp;
  lp->right = temp;
  rp->left = temp;

}


int hash_index( token t )
{
   int index;

//   index = ( int ) ( hash( t ) % G_HASH );
   index = ( int ) ( hash( t ) & G_HASH );
   return (index);
}


g_node_list insert_go( token t )
{

  g_node_list node;
  int index;

  if ( ( node = lookup_go( t ) ) != NULL ) return node;

  if ( gam_total_num_goals == MAX_MAX_NODES )
  {
    printf("\ngam: too many facts. increase MAX_MAX_NODES, currently %d\n\n", MAX_MAX_NODES);
    exit( 1 );
  }

  if ( debug == 7 )
    printf("\ninserting goal %s now\n", t);

  index = hash_index( t );

  node = ( g_node_list ) calloc( 1, sizeof( g_node_list_elt ) );
  CHECK_MEMORY( node );
  GRAPHF(g_node_list_elt);
  node->name = t;

  node->false_facts = NULL;
  node->true_facts = NULL;

  node->Tcount = 0;
  node->is_true = FALSE;
  node->st = 0;
  node->hits = 0;
  node->uid_block = 0;
  node->uid_mask = 0;
  node->is_initial = FALSE;
  node->old_selected = FALSE;
  node->new_selected = FALSE;
  node->count = 0;
  node->initial_count = 0;
  node->curr_true = FALSE;
  node->made_true = FALSE;
  node->made_false = FALSE;
  node->preconds = NULL;
  node->effconds = NULL;
  node->add_edges = NULL;
  node->del_edges = NULL;

  node->next = g_table[index];
  g_table[index] = node;

  set_gam_go_uid( node );
  gam_total_num_goals++;

  node->has_noop = FALSE;
  node->is_new = TRUE;

  return node;

}


g_node_list lookup_go( token t )
{

  g_node_list g;

  for ( g = g_table[ hash_index( t ) ]; g; g = g->next )
    if ( strcmp( g->name, t ) == SAME ) break;

  if ( g )
    return g;
  else
    return NULL;

}


o_node_list insert_op( token t )
{

  o_node_list node;
  int index;

  static int num = 0;

  if ( ( node = lookup_op( t ) ) != NULL ) return node;

  if ( num++ == MAX_MAX_NODES ) 
  {
    printf("\ngam: too many ops. increase MAX_MAX_NODES, currently %d\n\n", MAX_MAX_NODES);
    exit( 1 );
  }

  if ( debug == 7 )
    printf("\ninserting op %s now\n", t);

  index = hash_index( t );

  node = ( o_node_list ) calloc( 1, sizeof( o_node_list_elt ) );
  CHECK_MEMORY( node );GRAPHO(o_node_list_elt);
  node->name = t;

  node->preconds = NULL;
  node->effects = NULL;

  node->uid_block = 0;
  node->uid_mask = 0;

  node->next = o_table[index];
  o_table[index] = node;

  set_gam_op_uid( node );
  gam_total_num_ops++;

  node->is_noop = FALSE;

  return node;

}


o_node_list lookup_op( token t )
{

  o_node_list o;

  for ( o = o_table[ hash_index( t ) ]; o; o = o->next )
    if ( strcmp( o->name, t ) == SAME ) break;

  if ( o )
    return o;
  else
    return NULL;

}


BOOLEAN applicable( token_list list )
{

  token_list t;
  g_node_list g;

  for ( t = list; t; t = t->next )
    if ( ( g = lookup_go( t->item ) ) == NULL ) break;

  if ( t )
    return FALSE;
  else
    return TRUE;

}


g_edge_list insert_g_edge( g_node_list g, g_edge_list list )
{

  g_edge_list temp;

  temp = ( g_edge_list ) calloc( 1, sizeof( g_edge_list_elt ) );
  CHECK_MEMORY( temp );GRAPHK(g_edge_list_elt);
  temp->endpt = g;
  temp->next = list;

  return temp;

}


o_edge_list insert_o_edge( o_node_list o, o_edge_list list )
{

  o_edge_list temp;

  temp = ( o_edge_list ) calloc( 1, sizeof( o_edge_list_elt ) );
  CHECK_MEMORY( temp );GRAPHK(o_edge_list_elt);
  temp->endpt = o;
  temp->next = list;

  return temp;

}


e_edge_list insert_e_edge( e_node_list e, e_edge_list list )
{

  e_edge_list temp;

  temp = ( e_edge_list ) calloc( 1, sizeof( e_edge_list_elt ) );
  CHECK_MEMORY( temp );GRAPHK(e_edge_list_elt);
  temp->endpt = e;
  temp->next = list;

  return temp;

}


void print_ef_node( e_node_list e )
{

  g_edge_list g;

  printf( "con:\n" );
  for ( g = e->conditions; g; g = g->next )
    printf( "%s\n", g->endpt->name );
  printf( "add:\n" );
  for ( g = e->add_edges; g; g = g->next )
    printf( "%s\n", g->endpt->name );
  printf( "del:\n" );
  for ( g = e->del_edges; g; g = g->next )
    printf( "%s\n", g->endpt->name );

}


void print_go_op_nodes( void )
{

  o_node_list o;
  g_node_list g;
  e_node_list e;
  g_edge_list g_e;
  o_edge_list o_e;
  e_edge_list e_e;
  int i;

  for ( i = 0; i < G_HSIZE; i++ )
  {
    for ( o = o_table[i]; o; o = o->next ) 
    {
      printf( "operator %s:\n", o->name );
      printf( "pre:\n" );
      for ( g_e = o->preconds; g_e; g_e = g_e->next )
	printf( "%s\n", g_e->endpt->name );
      printf( "eff:\n" );
      for ( e = o->effects; e; e = e->next )
	print_ef_node( e );
    }
  }

  for ( i = 0; i < G_HSIZE; i++ )
    for ( g = g_table[i]; g; g = g->next ) 
    {
      printf( "goal: %s\npre:\n", g->name );
      for ( o_e = g->preconds; o_e; o_e = o_e->next )
	printf( "%s\n", o_e->endpt->name );
      printf( "add:\n" );
      for ( e_e = g->add_edges; e_e; e_e = e_e->next )
	print_ef_node( e_e->endpt );
      printf( "del:\n" );
      for ( e_e = g->del_edges; e_e; e_e = e_e->next )
	print_ef_node( e_e->endpt );
    }
}


BOOLEAN contain( g_edge_list l1, g_edge_list l2 )
{

  g_edge_list i, j;

  for ( i = l2; i; i = i->next ) 
  {
    for ( j = l1; j; j = j->next )
      if ( i->endpt == j->endpt ) break;
    if ( !j ) break;
  }

  if ( i )
    return FALSE;
  else
    return TRUE;

}


/* you should know about that one
 *
 * if you don't: look it up in
 *   T. Ottmann / P.Widmayer
 *   Algorithmen und Datenstrukturen, 3. Auflage
 *   page 85/86, Quicksort-Varianten, median of three quicksort
 */
void gam_quicksort( g_array_type a, int l, int r )
{

  int v, i, j, m;
  g_node_list t;

  if ( r > l )
  {
    m = ( int ) ( r + l ) / 2;
    if ( a[l]->st > a[r]->st ) 
    {
      t = a[l];
      a[l] = a[r];
      a[r] = t;
    }
    if ( a[l]->st > a[m]->st ) 
    {
      t = a[l];
      a[l] = a[m];
      a[m] = t;
    }
    if ( a[r]->st > a[m]->st ) 
    {
      t = a[r];
      a[r] = a[m];
      a[m] = t;
    }
    i = l - 1;
    j = r;
    v = a[r]->st;
    while( i < j ) 
    {
      do i++; while ( a[i]->st < v );
      do j--; while ( j >= 0 && a[j]->st > v );
      if ( i < j ) 
      {
        t = a[i];
        a[i] = a[j];
        a[j] = t;
      }
    }
    t = a[i];
    a[i] = a[r];
    a[r] = t;
    gam_quicksort( a, l, i - 1 );
    gam_quicksort( a, i + 1, r );
  }

}


void gam_set_quicksort( goal_set_list_array_type a, int l, int r )
{

  int v, i, j, m;
  goal_set_list t;

  if ( r > l ) 
  {
    m = ( int ) ( r + l ) / 2;
    if ( a[l]->st > a[r]->st ) 
    {
      t = a[l];
      a[l] = a[r];
      a[r] = t;
    }
    if ( a[l]->st > a[m]->st ) 
    {
      t = a[l];
      a[l] = a[m];
      a[m] = t;
    }
    if ( a[r]->st > a[m]->st ) 
    {
      t = a[r];
      a[r] = a[m];
      a[m] = t;
    }
    i = l - 1;
    j = r;
    v = a[r]->st;
    while( i < j ) 
    {
      do i++; while ( a[i]->st < v );
      do j--; while ( j >= 0 && a[j]->st > v );
      if ( i < j ) {
        t = a[i];
        a[i] = a[j];
        a[j] = t;
      }
    }
    t = a[i];
    a[i] = a[r];
    a[r] = t;
    gam_set_quicksort( a, l, i - 1 );
    gam_set_quicksort( a, i + 1, r );
  }

}


void set_gam_go_uid( g_node_list g )
{

  int id = gam_total_num_goals;

  g->uid_block = id / 32;
  g->uid_mask = 1 << ( id % 32 );

}


void set_gam_op_uid( o_node_list o )
{

  int id = gam_total_num_ops;

  o->uid_block = id / 32;
  o->uid_mask = 1 << ( id % 32 );

}


void print_goal_set( goal_set_list l )
{

  goal_list g;

  for ( g = l->set; g; g = g->next )
    printf( "%s ", g->name );
  printf( "\n" );

}


void print_goal_list( goal_list l )
{

  goal_list g;

  for ( g = l; g; g = g->next )
    printf( "%s ", g->name );
  printf( "\n" );

}


void print_O( void )
{

  o_node_list o;
  e_node_list e;
  int i;

  printf("\ncomplete o* is:\n");

  for ( i=0; i<G_HSIZE; i++ ) {
    for ( o = o_table[i]; o; o = o->next )
    {
      for ( e = o->effects; e; e = e->next ) 
      {
	if ( e->on ) 
	{
	  printf("\nfollowing effect of op %s\n", e->op->name );
	  print_ef_node( e );
	}
      }
    }
  }

  printf( "\n" );

}


BOOLEAN gam_are_there_non_exclusive( token_list facts, int time )
{

  static g_array_type ft;

  token_list i_token;
  int num = 0, i, j;

  for ( i_token = facts; i_token; i_token = i_token->next )
  {
    ft[num] = lookup_go( i_token->item );

    if ( !ft[num] || ft[num]->is_new )
    {
       if ( ft[num] )
          printf("\n Goal is_new : %s \n", ft[num]->name);
       else
          printf("\n Goal '%s' not found !\n", i_token->item);

       return FALSE;
    }

    num++;
  }

  for ( i = 0; i < num; i++ )
    for ( j = i + 1; j < num; j++ )
      if ( gam_go_are_mutex( ft[i], ft[j], time ) )
      {
         printf("\n Goals are mutex : %s %s \n", ft[i]->name, ft[j]->name);
         return FALSE;
      }

  return TRUE;

}


BOOLEAN gam_get_them_non_exclusive( token_list facts, int time, g_array_type *ft, int *num )
{

  token_list i_token;
  int i, j;

  *num = 0;

  for ( i_token = facts; i_token; i_token = i_token->next ) 
  {
    (*ft)[*num] = lookup_go( i_token->item );
    if ( !(*ft)[*num] || ((*ft)[*num])->is_new ) return FALSE;
    (*num)++;
  }

  for ( i = 0; i < *num; i++ )
    for ( j = i + 1; j < *num; j++ )
      if ( gam_go_are_mutex( (*ft)[i], (*ft)[j], time ) ) return FALSE;

  return TRUE;

}


e_node_list effect_is_already_there( o_node_list o, inst_effect_list e, int time )
{

  e_node_list i;
  g_edge_list g;
  token_list t;

  for ( i = o->effects; i; i = i->next ) 
  {

    for ( g = i->conditions; g; g = g->next ) 
    {
      for ( t = e->conditions; t; t = t->next )
	if ( strcmp( g->endpt->name, t->item ) == SAME ) break;

      if ( !t ) break;
    }
    if ( g ) continue;

    for ( t = e->conditions; t; t = t->next ) 
    {
      for ( g = i->conditions; g; g = g->next )
	if ( strcmp( g->endpt->name, t->item ) == SAME ) break;

      if ( !g ) break;
    }
    if ( t ) continue;

    return ( i );

  }

  return NULL;

}


BOOLEAN gam_go_are_mutex( g_node_list g1, g_node_list g2, int time )
{

  g_excl_list ex1 = g1->excl_at[time];

  if ( time == 0 )
     return FALSE;

  if ( !ex1 ) 
  {
    printf( "\ndidnt find goal %s excl vector at time %d\n", g1->name, time );
    exit( 1 );
  }

  return ( ( ( ex1->exclusive_vect[g2->uid_block] ) & ( g2->uid_mask ) ) > 0 );

}


void gam_go_make_exclusive( g_node_list g1, g_node_list g2, int time )
{

  g_excl_list ex1, ex2;

  if ( gam_go_are_mutex( g1, g2, time ) )
     return;

  ex1 = g1->excl_at[time];
  ex2 = g2->excl_at[time];

  if ( !ex1 ) 
  {
    printf( "\ndidnt find goal %s excl vector at time %d\n", g1->name, time );
    exit( 1 );
  }
  if ( !ex2 ) 
  {
    printf( "\ndidnt find goal %s excl vector at time %d\n", g2->name, time );
    exit( 1 );
  }

  if ( debug == 7 ) 
    printf("\ngam making exclusive %s and %s\n", g1->name, g2->name);

  ex1->exclusive_vect[g2->uid_block] |= g2->uid_mask;
  ex2->exclusive_vect[g1->uid_block] |= g1->uid_mask;

}


BOOLEAN gam_op_are_mutex( o_node_list o1, o_node_list o2, int time )
{

  o_excl_list ex1 = o1->excl_at[time];

  if ( !ex1 )
  {
    printf( "\ndidnt find op %s excl vector at time %d\n", o1->name, time );
    exit( 1 );
  }

  return ( ( ( ex1->exclusive_vect[o2->uid_block] ) & ( o2->uid_mask ) ) > 0 );

}


void gam_op_make_exclusive( o_node_list o1, o_node_list o2, int time )
{

  o_excl_list ex1, ex2;

  if ( gam_op_are_mutex( o1, o2, time ) )
     return;

  ex1 = o1->excl_at[time];
  ex2 = o2->excl_at[time];

  if ( !ex1 ) 
  {
    printf( "\ndidnt find op %s excl vector at time %d\n", o1->name, time );
    exit( 1 );
  }
  if ( !ex2 ) 
  {
    printf( "\ndidnt find op %s excl vector at time %d\n", o2->name, time );
    exit( 1 );
  }

  if ( debug == 7 )
    printf("\ngam making exclusive %s and %s\n", o1->name, o2->name);

  ex1->exclusive_vect[o2->uid_block] |= o2->uid_mask;
  ex2->exclusive_vect[o1->uid_block] |= o1->uid_mask;

}


BOOLEAN no_prev_time( g_node_list g, int time )
{
  return ( ( !g->excl_at[time-1] && time > 1 ) || ( !g->is_initial && time == 1 ) );
}


void print_F( g_node_list g )
{

  g_edge_list h;

  for ( h = g->false_facts; h; h = h->next )
    printf("%s\n", h->endpt->name);
}


void print_T( g_node_list g )
{

  g_edge_list h;

  for ( h = g->true_facts; h; h = h->next )
    printf("%s\n", h->endpt->name);
}



/* takes pointer lp to the left of the goal set that has to be
 * splitted in global list, similarly rp to the right;
 * i.e. lp<->critical_set<->rp;
 *
 * splits critical_set into several subsets and inserts them into
 * global list, resulting in
 * lp<->critical_subset_1<-> ... <->critical_subset_n<->rp
 */
void compute_agenda( goal_set_list lp, goal_set_list rp )
{

  static g_array_type goals;
  static BOOLEAN m[MAX_GOALS][MAX_GOALS];
  static goal_set_list_array_type goal_sets;

  goal_list goal, prev, curr = NULL, t_gl;
  int i, j, k, num = 0, status = -MAX_GOALS, num_sets = 0;
  goal_set_list t_gs, old = lp->right, un_ordered = NULL;

  BOOLEAN first = TRUE;
  token_list dummy;

  for ( goal = old->set; goal; goal = goal->next )
  {
    //#### MODIFICA GOAL AGENDA
    goal->num_unique = num;
    if ( ( goals[num] = lookup_go( goal->name ) ) == NULL ) break;
    else 
    {
      for ( i = 0; i < num; i++ )
	if ( gam_go_are_mutex( goals[i], goals[num], gam_fixpoint_level ) ) 
	{
	  printf("\ngam: goals are exclusive in fixpoint. problem unsolvable\n\n");
	  exit( 1 );
	}
      goals[num++]->st = 0;
    }
  }
  if ( goal ) 
  {
    printf( "\ngam: goal can't be reached: problem is unsolvable\n\n" );
    exit( 1 );
  }

  for ( i=0; i < num; i++ ) 
  {
    setup_FT( goals[i] );
    if ( debug == 123 ) 
    {
      printf("\ngam: false set of %s is:\n", goals[i]->name);
      print_F( goals[i] );
      printf("\n");
      printf("\ngam: true set of %s is:\n", goals[i]->name);
      print_T( goals[i] );
      printf("\n");
    }
    goals[i]->hits = 0;
  }

  for ( i = 0; i < num - 1; i++ ) 
  {
    setup_O( goals[i] );
    for ( j = i + 1; j < num; j++ ) 
    {
      if ( ( m[j][i] = !reachable( goals[j] ) ) == TRUE && display_info )
	printf("\nconstraining %s before %s", goals[j]->name, goals[i]->name);
    }
    unsetup_O( goals[i] );
  }
  
  for ( i = num-1; i > -1; i-- ) 
  {
    setup_O( goals[i] );
    for ( j = i - 1; j > -1; j-- ) 
    {
      if ( ( m[j][i] = !reachable( goals[j] ) ) == TRUE && display_info )
	printf("\nconstraining %s before %s", goals[j]->name, goals[i]->name);
    }
    unsetup_O( goals[i] );
  }

  for ( i = 0; i < num; i++ )
    m[i][i] = TRUE;

  for ( j = 0; j < num; j++ )
    for ( i = 0; i < num; i++ )
      if ( m[i][j] )
	for ( k = 0; k < num; k++ )
	  if ( m[j][k] )
	    m[i][k] = TRUE;

  for ( i = 0; i < num; i++ )
    for ( j = 0; j < num; j++ )
      if ( m[i][j] && i != j ) 
      {
	goals[i]->st--;
	goals[i]->hits++;
	goals[j]->st++;
	goals[j]->hits++;
      }

  gam_quicksort( goals, 0, num - 1 );

  for ( i = 0; i < num; i++ )
    if ( goals[i]->hits == 0 ) 
    {
      if ( debug )
	printf("\nunordered fact %s\n", goals[i]->name );
	
      t_gl = ( goal_list ) calloc( 1, sizeof( goal_list_elt ) );
      CHECK_MEMORY( t_gl );
      t_gl->name = goals[i]->name;
      dummy = gam_token_list_from_token( goals[i]->name );
      t_gl->objects = dummy->next;
      t_gl->next = NULL;
      for ( j = i + 1; j < num; j++ )
	goals[j - 1] = goals[j];
      num--;
      i--;
      if ( first ) 
      {
	t_gs = ( goal_set_list ) calloc( 1, sizeof( goal_set_list_elt ) );
	CHECK_MEMORY( t_gs );
	t_gs->intermediates = old->intermediates;
	t_gs->right = NULL;
	t_gs->left = NULL;
	t_gs->set = t_gl;
	curr = t_gs->set;
	first = FALSE;
	un_ordered = t_gs;
      } 
      else 
      {
	curr->next = t_gl;
	curr = t_gl;
      }
    }

  for ( i = 0; i < num; i++ ) 
  {
    if ( goals[i]->st != status ) 
    {
      if ( curr )
	curr->next = NULL;
      t_gs = ( goal_set_list ) calloc( 1, sizeof( goal_set_list_elt ) );
      CHECK_MEMORY( t_gs );
      t_gs->intermediates = old->intermediates;
      t_gs->right = NULL;
      t_gs->left = NULL;
      goal_sets[num_sets++] = t_gs;
      if ( strcmp( old->set->name, goals[i]->name ) == SAME ) 
      {
	t_gs->set = old->set;
	curr = old->set;
	old->set = old->set->next;
      } 
      else 
      {
	prev = old->set;
	for ( goal = old->set->next; goal; goal = goal->next )
	  if ( strcmp( goal->name, goals[i]->name ) == SAME ) break;
	  else prev = prev->next;
	if ( !goal && debug )
	  printf("didn't find goal where it ought to be. debug me, please");
	t_gs->set = goal;
	prev->next = goal->next;
	curr = goal;
      }
      status = goals[i]->st;
    } 
    else 
    {
      if ( strcmp( old->set->name, goals[i]->name ) == SAME ) 
      {
	curr->next = old->set;
	old->set = old->set->next;
	curr = curr->next;
      } 
      else 
      {
	prev = old->set;
	for ( goal = old->set->next; goal; goal = goal->next )
	  if ( strcmp( goal->name, goals[i]->name ) == SAME ) break;
	  else prev = prev->next;
	if ( !goal && debug )
	  printf("didn't find goal where it ought to be. debug me, please");
	curr->next = goal;
	prev->next = goal->next;
	curr = curr->next;
      }
    }
  }
  if ( curr )
    curr->next = NULL;

  if ( un_ordered ) 
  {
    goal_sets[num_sets++] = un_ordered;

    for ( i = 0; i < num_sets; i++ )
      for ( j = 0; j < num_sets; j++ )
	m[i][j] = i > j ? FALSE : TRUE;

    for ( i = 0; i < num_sets - 1; i++ )
      m[num_sets - 1][i] =
	!set_reachable( goal_sets[i], goal_sets[num_sets - 1] );
    for ( i = 0; i < num_sets - 1; i++ )
      m[i][num_sets - 1] =
	!set_reachable( goal_sets[num_sets - 1], goal_sets[i] );

    for ( j = 0; j < num_sets; j++ )
      for ( i = 0; i < num_sets; i++ )
	if ( m[i][j] )
	  for ( k = 0; k < num_sets; k++ )
	    if ( m[j][k] )
	      m[i][k] = TRUE;

    for ( i = 0; i < num_sets; i++ ) 
    {
      goal_sets[i]->st = 0;
      goal_sets[i]->hits = 0;
    }
    for ( i = 0; i < num_sets; i++ )
      for ( j = 0; j < num_sets; j++ )
	if ( m[i][j] && i != j ) 
	{
	  goal_sets[i]->st--;
	  goal_sets[i]->hits++;
	  goal_sets[j]->st++;
	  goal_sets[j]->hits++;
	}

    for ( i = 0; i < num_sets; i++ )
      if ( goal_sets[i]->hits == 0 ) 
      {
	if ( debug )
	  printf( "\nfound unorderable goal set\n" );
	for ( goal = old->set; goal->next; goal = goal->next );
	curr = goal;
	for ( i = 0; i < num_sets - 1; i++ ) 
	{
	  curr->next = goal_sets[i]->set;
	  for ( goal = goal_sets[i]->set; goal->next; goal = goal->next );
	  curr = goal;
	}
	return;
      }

    gam_set_quicksort( goal_sets, 0, num_sets - 1 );

  } 
  else
    for ( i = 0; i < num_sets; i++ ) goal_sets[i]->st = i;

  status = -MAX_GOALS;
  for ( i = 0; i < num_sets; i++ ) 
  {
    if ( goal_sets[i]->st > status ) 
    {
      lp->right = goal_sets[i];
      goal_sets[i]->left = lp;
      lp = goal_sets[i];
      status = goal_sets[i]->st;
    } 
    else 
    {
      for ( goal = lp->set; goal->next; goal = goal->next );
      goal->next = goal_sets[i]->set;
    }
  }
  rp->left = lp;
  lp->right = rp;

  // initialise goal orderings vector ...
  for (i=0; i<num; i++)
    for (j=0; j<num; j++)
    {
       if ( m[i][j] != m[j][i] )
          GLOBAL_first_before_second[i][j] = m[i][j];
       else
          GLOBAL_first_before_second[i][j] = FALSE;
    }

}


BOOLEAN set_reachable( goal_set_list l1, goal_set_list l2 )
{

  BOOLEAN result = TRUE;

  static g_array_type g1;

  int i, num = 0;
  goal_list go1, go2;


  if ( debug ) 
  {
    printf ( "\ntry constraining sets:\n");
    print_goal_set( l1 );
    print_goal_set( l2 );
  }

  for ( go1 = l1->set; go1; go1 = go1->next )
    g1[num++] = lookup_go( go1->name );

  for ( i = 0; i < num; i++ )
    setup_O( g1[i] );

  if ( debug == 1 )
    print_O();

  for ( go2 = l2->set; go2; go2 = go2->next ) 
  {
    if ( !reachable( lookup_go( go2->name ) ) ) 
    {
      if ( debug ) 
      {
	printf ( "\nconstraining sets:\n");
	print_goal_set( l1 );
	print_goal_set( l2 );
      }
      result = FALSE;
      break;
    }
  }

  for ( i = 0; i < num; i++ )
    unsetup_O( g1[i] );

  return result;

}


void setup_FT( g_node_list g1 )
{

  int i, Tnum = 0;
  g_node_list g;
  o_node_list o;
  e_node_list e, e_;
  e_edge_list e_e;
  g_edge_list g_e;

  if ( debug == 2 )
    printf ("\nset up T for %s\n", g1->name );

  for ( i = 0; i < G_HSIZE; i++ )
    for ( g = g_table[i]; g; g = g->next ) 
    {
      g->Tcount = 0;
      if ( g1 != g && gam_go_are_mutex( g1, g, gam_fixpoint_level ) ) 
      {
	if ( 0 )
	  printf("false set of %s contains %s\n", g1->name, g->name);
	g1->false_facts = insert_g_edge( g, g1->false_facts );
      }
    }

  if ( !do_true_set ) return;

  for ( e_e = g1->add_edges; e_e; e_e = e_e->next ) 
  {
    e = e_e->endpt;
    Tnum++;
    for ( g_e = e->add_edges; g_e; g_e = g_e->next )
      g_e->endpt->Tcount++;
    o = e->op;
    for ( e_ = o->effects; e_; e_ = e_->next ) 
    {
      if ( contain( e->conditions, e_->conditions ) ) 
      {
	for ( g_e = e_->del_edges; g_e; g_e = g_e->next )
	  g_e->endpt->Fcount++;
      }
    }
  }

  for ( i = 0; i < G_HSIZE; i++ )
    for ( g = g_table[i]; g; g = g->next ) 
    {
      if ( ( g->is_initial || g->Tcount == Tnum ) && g->Fcount == 0 ) 
      {
	g1->true_facts = insert_g_edge( g, g1->true_facts );
      }
    }

}


BOOLEAN reachable( g_node_list g )
{

  e_edge_list e_e;
  g_edge_list g_e;

  if ( 0 && g->is_true ) return TRUE;

  for ( e_e = g->add_edges; e_e; e_e = e_e->next ) 
  {
    if ( debug == 1 )
      printf("\ntrying add-op for %s: %s now", g->name, e_e->endpt->op->name );
    if ( !e_e->endpt->on ) continue;
    for ( g_e = e_e->endpt->op->preconds; g_e; g_e = g_e->next )
      if ( !is_added( g_e->endpt ) ) break;
    if ( g_e ) continue;
    for ( g_e = e_e->endpt->conditions; g_e; g_e = g_e->next )
      if ( !is_added( g_e->endpt ) ) break;
    if ( g_e ) continue;
    if ( debug == 1 )
      printf("managed");

    return TRUE;
  }

  return FALSE;

}


BOOLEAN is_added( g_node_list g )
{

  e_edge_list e_e;

  if ( g->is_true ) 
  {
    if ( debug == 1 )
      printf("\ngoal %s is true already", g->name );
    return TRUE;
  }

  for ( e_e = g->add_edges; e_e; e_e = e_e->next ) 
  {
    if ( !e_e->endpt->on ) continue;
    return TRUE;
  }

  if ( debug == 1 )
    printf( "\nfound not added fact %s", g->name );

  return FALSE;

}


void setup_O( g_node_list g1 )
{

  g_edge_list g_e;
  o_edge_list o_e;
  e_node_list e;
  e_edge_list e_e;

  for ( g_e = g1->true_facts; g_e; g_e = g_e->next )
    g_e->endpt->is_true = TRUE;

  for ( g_e = g1->false_facts; g_e; g_e = g_e->next ) 
  {
    for ( o_e = g_e->endpt->preconds; o_e; o_e = o_e->next ) 
    {
      if ( o_e->endpt->is_noop ) continue;
      for ( e = o_e->endpt->effects; e; e = e->next )
	e->on = FALSE;
    }
    for ( e_e = g_e->endpt->effconds; e_e; e_e = e_e->next ) 
    {
      e_e->endpt->on = FALSE;
    }

  }

  for ( e_e = g1->del_edges; e_e; e_e = e_e->next ) 
  {
    if ( !e_e->endpt->conditions ) 
    {
      for ( e = e_e->endpt->op->effects; e; e = e->next )
	e->on = FALSE;
    } 
    else 
      e_e->endpt->on = FALSE;
  }

}


void unsetup_O( g_node_list g1 )
{

  g_edge_list g_e;
  o_edge_list o_e;
  e_node_list e;
  e_edge_list e_e;

  for ( g_e = g1->true_facts; g_e; g_e = g_e->next ) 
  {
    g_e->endpt->is_true = FALSE;
  }

  for ( g_e = g1->false_facts; g_e; g_e = g_e->next ) 
  {
    for ( o_e = g_e->endpt->preconds; o_e; o_e = o_e->next ) 
    {
      if ( o_e->endpt->is_noop ) continue;
      for ( e = o_e->endpt->effects; e; e = e->next )
	e->on = TRUE;
    }
    for ( e_e = g_e->endpt->effconds; e_e; e_e = e_e->next ) 
    {
      e_e->endpt->on = TRUE;
    }
  }

  for ( e_e = g1->del_edges; e_e; e_e = e_e->next ) 
  {
    if ( !e_e->endpt->conditions ) 
    {
      for ( e = e_e->endpt->op->effects; e; e = e->next )
	e->on = TRUE;
    } 
    else
      e_e->endpt->on = TRUE;
  }
}


goal_list intermediate( goal_list goals )
{

  int i;
  goal_list result, go;
  bit_string result_string;
  g_node_list g;

  if ( debug == 1 ) 
  {
    printf("\nnow computing intermediates for:\n");
    for ( go = goals; go; go = go ->next )
      printf( "%s ", go->name );
  }

  for ( i = 0; i < G_HSIZE; i++ )
    for ( g = g_table[i]; g; g = g->next )
      g->count = 0;

  for ( i = 0; i < NUMINTS; i++ ) result_string[i] = 0;

  for ( go = goals; go; go = go->next ) 
  {
    for ( i = 0; i < G_HSIZE; i++ )
      for ( g = g_table[i]; g; g = g->next ) 
      {
	g->old_selected = FALSE;
	g->new_selected = FALSE;
      }
    g = lookup_go( go->name );
    g->old_selected = TRUE;
    generate_intersections();
    if ( debug ) 
    {
      printf( "\ninters for %s:\n", go->name );
      print_goal_list( make_goal_list( temp_string ) );
    }
    for ( i = 0; i < NUMINTS; i++ ) result_string[i] |= temp_string[i];
  }

  result = make_goal_list( result_string );

  if ( result && debug ) 
  {
    printf( "\ndiscovered intermediates:\n" );
    for ( go = result; go; go = go->next )
      printf( "%s ", go->name );
    printf( "\n\n" );
  }

  return result;

}


void generate_intersections( void )
{

  int i, j, num;
  bit_string_list curr = first_string->next, temp;
  e_edge_list e;
  g_node_list g, go;
  g_edge_list g_e;

  while ( TRUE ) 
  {
    if ( curr == last_string ) 
    {
      temp = ( bit_string_list ) calloc( 1, sizeof( bit_string_list_elt ) );
      temp->prev = curr->prev;
      curr->prev->next = temp;
      temp->next = last_string;
      last_string->prev = temp;
      for ( i = 0; i < NUMINTS; i++ ) 
         (temp->string)[i] = 0;
      curr = temp;
    } 
    else
    {
      for ( i = 0; i < NUMINTS; i++ ) 
         (curr->string)[i] = 0;
    }
    
    for ( i = 0; i < G_HSIZE; i++ ) 
    {
      for ( go = g_table[i]; go; go = go->next ) 
      {
	if ( !go->old_selected ) continue;
	num = 0;
	for ( e = go->add_edges; e; e = e->next ) 
	{
	  num++;
	  for ( g_e = e->endpt->op->preconds; g_e; g_e = g_e->next )
	    g_e->endpt->count++;
	  for ( g_e = e->endpt->conditions; g_e; g_e = g_e->next )
	    g_e->endpt->count++;
	}
	for ( j = 0 ; j < G_HSIZE; j++ ) 
	{
	  for ( g = g_table[j]; g; g = g->next ) 
	  {
	    if ( ( g->count == num ) && ( !g->is_initial ) ) 
	    {
	      g->new_selected = TRUE;
	      if ( debug == 1 )
		printf( "\njust selected %s\n", g->name );
	      (curr->string)[g->uid_block] |= g->uid_mask;
	    }
	    g->count = 0;
	  }
	}
      }
    }

    for ( i = 0; i < G_HSIZE; i++ )
      for ( go = g_table[i]; go; go = go->next ) 
      {
	go->old_selected = go->new_selected;
	go->new_selected = FALSE;
      }	

    for ( i = 0; i < NUMINTS; i++ )
      if ( (curr->string)[i] != 0 ) break;
    if ( i == NUMINTS ) break;

    if ( circle_occured( curr ) ) break;

    curr = curr->next;

  }

  if ( curr->prev == first_string ) 
  {
    for ( i = 0; i < NUMINTS; i++ ) 
       temp_string[i] = 0;
  } 
  else 
  {
    for ( i = 0; i < NUMINTS; i++ ) 
       temp_string[i] = (curr->prev->string)[i];
  }

}


BOOLEAN circle_occured( bit_string_list curr )
{

  bit_string_list i;
  int j;

  for ( i = curr->prev; i != first_string; i = i->prev ) 
  {
    for ( j = 0; j < NUMINTS; j++ )
      if ( (curr->string)[j] != (i->string)[j] ) break;
    if ( j == NUMINTS ) return TRUE;
  }

  return FALSE;

}


goal_list make_goal_list( bit_string vec )
{

  int i;
  g_node_list g;
  goal_list result = NULL, temp;
  token_list dummy;

  for ( i = 0; i < G_HSIZE; i++ )
    for ( g = g_table[i]; g; g = g->next )
      if ( ( ( vec[g->uid_block] ) & ( g->uid_mask ) ) > 0 ) 
      {
	temp = ( goal_list ) calloc( 1, sizeof( goal_list_elt ) );
	temp->name = g->name;
	dummy = gam_token_list_from_token( g->name );
	temp->objects = dummy->next;
	temp->next = result;
	result = temp;
      }

  return result;

}
